

import pytplot
from pyspedas import tnames

def tkm2re(name, km=False, newname=None, suffix=''):
    """
    Converts a tplot variable to Re or Km

    Input
    ------
        name: str or list of str
            Names of tplot variables to convert

    Parameters
    -----------
        km: bool
            Flag to convert Re to Km

        newname: str or list of str
            Output variable names; if not set, the names of the
            input variables are used + '_re' or '_km'

        suffix: str
            Specify to append a suffix to each variable 
            (only valid if newname is not specified)

    Returns
    --------
        List of the tplot variables created

    """
    km_in_re = 6371.2

    names = tnames(name)

    if names == []:
        print('No tplot variables found: ' + name)
        return

    if newname is None:
        newname = [n + suffix for n in names]

        if km == False:
            newname = [n + '_re' for n in newname]
        else:
            newname = [n + '_km' for n in newname]
    else:
        if not isinstance(newname, list):
            newname = [newname]

        if len(newname) != len(names):
            print('Number of output variable names (newname) should match the number of input variables.')
            return

    out = []

    for in_tvar, out_tvar in zip(names, newname):
        data = pytplot.get_data(in_tvar)
        metadata = pytplot.get_data(in_tvar, metadata=True)

        if data is None:
            print('Problem reading variable: ' + in_tvar)
            continue

        if km == False:
            data_out = data.y/km_in_re
        else:
            data_out = data.y*km_in_re

        saved = pytplot.store_data(out_tvar, data={'x': data.times, 'y': data_out}, attr_dict=metadata)

        if not saved:
            print('Problem creating tplot variable.')
            continue

        # update the subtitle, if needed
        yaxis_opt = pytplot.data_quants[out_tvar].attrs['plot_options'].get('yaxis_opt')

        if yaxis_opt is not None:
            subtitle = yaxis_opt.get('axis_subtitle')
            if subtitle is not None:
                if km == False:
                    new_subtitle = pytplot.data_quants[out_tvar].attrs['plot_options']['yaxis_opt']['axis_subtitle'].lower().replace('km', 'Re')
                else:
                    new_subtitle = pytplot.data_quants[out_tvar].attrs['plot_options']['yaxis_opt']['axis_subtitle'].lower().replace('re', 'km')
                pytplot.data_quants[out_tvar].attrs['plot_options']['yaxis_opt']['axis_subtitle'] = new_subtitle

        out.append(out_tvar)

    return out


import pyspedas
import unittest

from pyspedas.utilities.data_exists import data_exists
from pyspedas.mms.fpi.mms_fpi_split_tensor import mms_fpi_split_tensor

class FPITestCases(unittest.TestCase):
    def test_split_tensors(self):
        data = pyspedas.mms.fpi(trange=['2015-10-16/13:06', '2015-10-16/13:07'],
                                data_rate='brst',
                                datatype=['dis-moms', 'des-moms'])

        split_vars = mms_fpi_split_tensor('mms1_dis_temptensor_gse_brst')

        components = ['xx', 'xy', 'xz', 'yx', 'yy', 'yz', 'zx', 'zy', 'zz']
        output = ['mms1_dis_temptensor_gse_brst_' + component for component in components]

        for v in output:
            self.assertTrue(data_exists(v))

        split_vars = mms_fpi_split_tensor('mms1_des_temptensor_gse_brst')

        components = ['xx', 'xy', 'xz', 'yx', 'yy', 'yz', 'zx', 'zy', 'zz']
        output = ['mms1_des_temptensor_gse_brst_' + component for component in components]

        for v in output:
            self.assertTrue(data_exists(v))



import unittest
import pyspedas
from pyspedas.utilities.data_exists import data_exists
from pyspedas.mms.fgm.mms_lingradest import mms_lingradest

class CurlTestCases(unittest.TestCase):
    def test_default(self):
        data = pyspedas.mms.fgm(probe=[1, 2, 3, 4], trange=['2015-10-30/05:15:45', '2015-10-30/05:15:48'], data_rate='brst', time_clip=True, varformat='*_gse_*', get_fgm_ephemeris=True)
        curl = pyspedas.mms.curlometer(positions=['mms1_fgm_r_gse_brst_l2', 'mms2_fgm_r_gse_brst_l2', 'mms3_fgm_r_gse_brst_l2', 'mms4_fgm_r_gse_brst_l2'], fields=['mms1_fgm_b_gse_brst_l2', 'mms2_fgm_b_gse_brst_l2', 'mms3_fgm_b_gse_brst_l2', 'mms4_fgm_b_gse_brst_l2'])
        self.assertTrue(data_exists('jtotal'))
        self.assertTrue(data_exists('curlB'))
        self.assertTrue(data_exists('divB'))
        self.assertTrue(data_exists('jpar'))
        self.assertTrue(data_exists('jperp'))

    def test_lingradest(self):
        data = pyspedas.mms.fgm(probe=[1, 2, 3, 4], trange=['2015-10-30/05:15:45', '2015-10-30/05:15:48'], data_rate='brst', time_clip=True, varformat='*_gse_*', get_fgm_ephemeris=True)
        fields = ['mms'+prb+'_fgm_b_gse_brst_l2' for prb in ['1', '2', '3', '4']]
        positions = ['mms'+prb+'_fgm_r_gse_brst_l2' for prb in ['1', '2', '3', '4']]
        curl = mms_lingradest(fields=fields, positions=positions, suffix='_lingradest')
        self.assertTrue(data_exists('Bx_lingradest'))
        self.assertTrue(data_exists('By_lingradest'))
        self.assertTrue(data_exists('Bz_lingradest'))
        self.assertTrue(data_exists('gradBx_lingradest'))
        self.assertTrue(data_exists('gradBy_lingradest'))
        self.assertTrue(data_exists('gradBz_lingradest'))
        self.assertTrue(data_exists('jx_lingradest'))
        self.assertTrue(data_exists('jy_lingradest'))
        self.assertTrue(data_exists('jz_lingradest'))
        self.assertTrue(data_exists('absCB_lingradest'))
        self.assertTrue(data_exists('CxB_lingradest'))
        self.assertTrue(data_exists('CyB_lingradest'))
        self.assertTrue(data_exists('CzB_lingradest'))
        self.assertTrue(data_exists('curvx_lingradest'))
        self.assertTrue(data_exists('curvy_lingradest'))
        self.assertTrue(data_exists('curvz_lingradest'))

    def test_lingradest_wrapper(self):
        data = pyspedas.mms.fgm(probe=[1, 2, 3, 4], trange=['2015-10-30/05:15:45', '2015-10-30/05:15:48'], data_rate='brst', time_clip=True, varformat='*_gse_*', get_fgm_ephemeris=True)
        fields = ['mms'+prb+'_fgm_b_gse_brst_l2' for prb in ['1', '2', '3', '4']]
        positions = ['mms'+prb+'_fgm_r_gse_brst_l2' for prb in ['1', '2', '3', '4']]
        curl = pyspedas.mms.lingradest(fields=fields, positions=positions, suffix='_lingradest')
        self.assertTrue(data_exists('Bx_lingradest'))
        self.assertTrue(data_exists('By_lingradest'))
        self.assertTrue(data_exists('Bz_lingradest'))
        self.assertTrue(data_exists('gradBx_lingradest'))
        self.assertTrue(data_exists('gradBy_lingradest'))
        self.assertTrue(data_exists('gradBz_lingradest'))
        self.assertTrue(data_exists('jx_lingradest'))
        self.assertTrue(data_exists('jy_lingradest'))
        self.assertTrue(data_exists('jz_lingradest'))
        self.assertTrue(data_exists('absCB_lingradest'))
        self.assertTrue(data_exists('CxB_lingradest'))
        self.assertTrue(data_exists('CyB_lingradest'))
        self.assertTrue(data_exists('CzB_lingradest'))
        self.assertTrue(data_exists('curvx_lingradest'))
        self.assertTrue(data_exists('curvy_lingradest'))
        self.assertTrue(data_exists('curvz_lingradest'))

if __name__ == '__main__':
    unittest.main()



import numpy as np
from pytplot import get_data, store_data
from geopack import geopack, t89

def tt89(pos_var_gsm, iopt=3, suffix='', igrf_only=False):
    """
    tplot wrapper for the functional interface to Sheng Tian's implementation 
    of the Tsyganenko 96 and IGRF model:

    https://github.com/tsssss/geopack

    Input
    ------
        pos_gsm_tvar: str
            tplot variable containing the position data (km) in GSM coordinates

    Parameters
    -----------
        iopt: int
            Specifies the ground disturbance level:
                iopt= 1       2        3        4        5        6      7
                           correspond to:
                kp=  0,0+  1-,1,1+  2-,2,2+  3-,3,3+  4-,4,4+  5-,5,5+  &gt =6-

        suffix: str
            Suffix to append to the tplot output variable

    Returns
    --------
        Name of the tplot variable containing the model data
    """
    pos_data = get_data(pos_var_gsm)

    if pos_data is None:
        print('Variable not found: ' + pos_var_gsm)
        return

    b0gsm = np.zeros((len(pos_data.times), 3))
    dbgsm = np.zeros((len(pos_data.times), 3))

    # convert to Re
    pos_re = pos_data.y/6371.2

    for idx, time in enumerate(pos_data.times):
        tilt = geopack.recalc(time)

        # dipole B in GSM
        b0gsm[idx, 0], b0gsm[idx, 1], b0gsm[idx, 2] = geopack.dip(pos_re[idx, 0], pos_re[idx, 1], pos_re[idx, 2])

        # T89 dB in GSM
        dbgsm[idx, 0], dbgsm[idx, 1], dbgsm[idx, 2] = t89.t89(iopt, tilt, pos_re[idx, 0], pos_re[idx, 1], pos_re[idx, 2])

    if igrf_only:
        bgsm = b0gsm
    else:
        bgsm = b0gsm + dbgsm

    saved = store_data(pos_var_gsm + '_bt89' + suffix, data={'x': pos_data.times, 'y': bgsm})

    if saved:
        return pos_var_gsm + '_bt89' + suffix


from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.analysis.time_clip import time_clip as tclip
from pytplot import cdf_to_tplot

from .config import CONFIG

def load(trange=['2018-11-5', '2018-11-6'], 
         probe='a',
         instrument='emfisis', 
         level='l3',
         datatype='magnetometer', 
         suffix='', 
         cadence='4sec', # for EMFISIS mag data
         coord='sm', # for EMFISIS mag data
         wavetype='waveform', # for EMFISIS waveform data
         rel='rel04', # for ECT data
         get_support_data=False, 
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False):
    """
    This function loads Van Allen Probes (RBSP) data; this function is not meant 
    to be called directly; instead, see the wrappers:
        pyspedas.rbsp.emfisis
        pyspedas.rbsp.rbspice
        pyspedas.rbsp.efw
        pyspedas.rbsp.mageis
        pyspedas.rbsp.hope
        pyspedas.rbsp.rept
        pyspedas.rbsp.rps
    
    """

    if not isinstance(probe, list):
        probe = [probe]

    out_files = []

    for prb in probe:
        if instrument == 'emfisis':
            if datatype == 'density' or datatype == 'housekeeping' or datatype == 'wna-survey':
                pathformat = 'rbsp'+prb+'/'+level+'/'+instrument+'/'+datatype+'/%Y/rbsp-'+prb+'_'+datatype+'_'+instrument+'-'+level+'_%Y%m%d_v*.cdf'
            elif datatype == 'wfr' or datatype == 'hfr':
                pathformat = 'rbsp'+prb+'/'+level+'/'+instrument+'/'+datatype+'/'+wavetype+'/%Y/rbsp-'+prb+'_'+datatype+'-'+wavetype+'_'+instrument+'-'+level+'_%Y%m%d*_v*.cdf'
            else:
                if level == 'l2' and datatype == 'magnetometer':
                    pathformat = 'rbsp'+prb+'/'+level+'/'+instrument+'/'+datatype+'/uvw/%Y/rbsp-'+prb+'_'+datatype+'_uvw_'+instrument+'-'+level+'_%Y%m%d*_v*.cdf'
                else:
                    pathformat = 'rbsp'+prb+'/'+level+'/'+instrument+'/'+datatype+'/'+cadence+'/'+coord+'/%Y/rbsp-'+prb+'_'+datatype+'_'+cadence+'-'+coord+'_'+instrument+'-'+level+'_%Y%m%d_v*.cdf'
        elif instrument == 'rbspice':
            pathformat = 'rbsp'+prb+'/'+level+'/'+instrument+'/'+datatype+'/%Y/rbsp-'+prb+'-'+instrument+'_lev-'+str(level[-1])+'?'+datatype+'_%Y%m%d_v*.cdf'
        elif instrument == 'efw':
            if level == 'l3':
                pathformat = 'rbsp'+prb+'/'+level+'/'+instrument+'/%Y/rbsp'+prb+'_'+instrument+'-'+level+'_%Y%m%d_v??.cdf'
            else:
                pathformat = 'rbsp'+prb+'/'+level+'/'+instrument+'/'+datatype+'/%Y/rbsp'+prb+'_'+instrument+'-'+level+'_'+datatype+'_%Y%m%d_v??.cdf'
        elif instrument == 'mageis':
            pathformat = 'rbsp'+prb+'/'+level+'/ect/'+instrument+'/sectors/'+rel+'/%Y/rbsp'+prb+'_'+rel+'_ect-mageis-'+level+'_%Y%m%d_v*.cdf'
        elif instrument == 'hope':
            if datatype == 'moments':
                pathformat = 'rbsp'+prb+'/'+level+'/ect/'+instrument+'/'+datatype+'/'+rel+'/%Y/rbsp'+prb+'_'+rel+'_ect-hope-mom-'+level+'_%Y%m%d_v*.cdf'
            elif datatype == 'pitchangle':
                pathformat = 'rbsp'+prb+'/'+level+'/ect/'+instrument+'/'+datatype+'/'+rel+'/%Y/rbsp'+prb+'_'+rel+'_ect-hope-pa-'+level+'_%Y%m%d_v*.cdf'
            elif datatype == 'spinaverage':
                pathformat = 'rbsp'+prb+'/'+level+'/ect/'+instrument+'/'+datatype+'/'+rel+'/%Y/rbsp'+prb+'_'+rel+'_ect-hope-sci-'+level+'sa_%Y%m%d_v*.cdf'
        elif instrument == 'rept':
            pathformat = 'rbsp'+prb+'/'+level+'/ect/'+instrument+'/sectors/'+rel+'/%Y/rbsp'+prb+'_'+rel+'_ect-rept-sci-'+level+'_%Y%m%d_v*.cdf'
        elif instrument == 'rps':
            if datatype == 'rps-1min':
                pathformat = 'rbsp'+prb+'/'+level+'/rps/psbr-rps-1min/%Y/rbsp'+prb+'_'+level+'-1min_psbr-rps_%Y%m%d_v*.cdf'
            elif datatype == 'rps':
                pathformat = 'rbsp'+prb+'/'+level+'/rps/psbr-rps/%Y/rbsp'+prb+'_'+level+'_psbr-rps_%Y%m%d_v*.cdf'


        # find the full remote path names using the trange
        remote_names = dailynames(file_format=pathformat, trange=trange)

        files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update)
        if files is not None:
            for file in files:
                out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars





from .load import load

def threedp(trange=['1999-11-5', '1999-11-6'],
        datatype='3dp_emfits_e0',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads 3DP data
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='3dp', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

def orbit(trange=['1999-11-5', '1999-11-6'],
        datatype='pre_or',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads orbit data
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='orbit', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

def sms(trange=['1999-11-5', '1999-11-6'],
        datatype='k0',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Solar Wind and Suprathermal Ion Composition Instrument
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='sms', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

def waves(trange=['2018-11-5', '2018-11-6'],
        datatype='h1',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads Radio/Plasma Wave (WAVES) data
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='waves', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

def mfi(trange=['2018-11-5', '2018-11-6'],
        datatype='h0',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the Fluxgate Magnetometer
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='fgm', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

def swe(trange=['2018-11-5', '2018-11-6'],
        datatype='h5',
        suffix='',  
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False):
    """
    This function loads data from the SWE instrument
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='swe', trange=trange, datatype=datatype, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)




import unittest
from pyspedas.utilities.data_exists import data_exists

import pyspedas

class LoadTestCases(unittest.TestCase):
    def test_load_dst_data(self):
        dst_vars = pyspedas.kyoto.dst(trange=['2015-10-15', '2015-10-16'])
        self.assertTrue(data_exists('kyoto_dst'))

if __name__ == '__main__':
    unittest.main()


import logging
from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.analysis.time_clip import time_clip as tclip
from pytplot import cdf_to_tplot

from .config import CONFIG


def load(trange=['2013-11-5', '2013-11-6'],
         instrument='fgm',
         probe='c',
         level='l2',
         datatype=None, # ASK data
         stations=None,  # ground mag and ASK data
         greenland=None,  # also for ground mag data
         suffix='',
         get_support_data=False,
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False):
    """
    This function loads data from the THEMIS mission;
    this function is not meant to be called directly;
    instead, see the wrappers:
        pyspedas.themis.fgm
        pyspedas.themis.fit
        pyspedas.themis.efi
        pyspedas.themis.scm
        pyspedas.themis.fft
        pyspedas.themis.fbk
        pyspedas.themis.esa
        pyspedas.themis.sst
        pyspedas.themis.mom
        pyspedas.themis.gmom
        pyspedas.themis.gmag
        pyspedas.themis.ask
        pyspedas.themis.state
        pyspedas.themis.slp

    """

    if not isinstance(probe, list):
        probe = [probe]

    out_files = []
    file_resolution = 24*3600.0 # default to daily files

    for prb in probe:
        if instrument == 'ask':
            if stations is None:
                pathformat = ('thg/' + level + '/asi/ask/%Y/thg_' + level + '_ask'
                              + '_%Y%m%d_v01.cdf')
            else:
                # individual station data should have hourly files
                file_resolution = 3600.0
                pathformat = ('thg/' + level + '/asi/' + stations + '/%Y/%m/'
                              + 'thg_' + level + '_' + datatype + '_' + stations
                              + '_%Y%m%d%H_v01.cdf')
        elif instrument == 'fgm':
            pathformat = ('th' + prb + '/' + level + '/' + instrument
                          + '/%Y/th' + prb + '_' + level + '_' + instrument
                          + '_%Y%m%d_v??.cdf')
        elif instrument == 'fit':
            pathformat = ('th' + prb + '/' + level + '/' + instrument
                          + '/%Y/th' + prb + '_' + level + '_' + instrument
                          + '_%Y%m%d_v??.cdf')
        elif instrument == 'efi':
            if level == 'l2':
                pathformat = ('th' + prb + '/' + level + '/' + instrument
                              + '/%Y/th' + prb + '_' + level + '_' + instrument
                              + '_%Y%m%d_v??.cdf')
            elif level == 'l1':
                pathformat = [('th' + prb + '/' + level + '/vaf/%Y/th' + prb
                               + '_' + level + '_vaf_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/vap/%Y/th' + prb
                               + '_' + level + '_vap_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/vaw/%Y/th' + prb
                               + '_' + level + '_vaw_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/vbf/%Y/th' + prb
                               + '_' + level + '_vbf_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/vbp/%Y/th' + prb
                               + '_' + level + '_vbp_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/vbw/%Y/th' + prb
                               + '_' + level + '_vbw_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/eff/%Y/th' + prb
                              + '_' + level + '_eff_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/efw/%Y/th' + prb
                               + '_' + level + '_efw_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/efp/%Y/th' + prb
                               + '_' + level + '_efp_%Y%m%d_v??.cdf')]
        elif instrument == 'scm':
            if level == 'l2':
                pathformat = ('th' + prb + '/' + level + '/' + instrument
                              + '/%Y/th' + prb + '_' + level + '_' + instrument
                              + '_%Y%m%d_v??.cdf')
            elif level == 'l1':
                pathformat = [('th' + prb + '/' + level + '/scp/%Y/th' + prb
                               + '_' + level + '_scp_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/scf/%Y/th' + prb
                               + '_' + level + '_scf_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/scw/%Y/th' + prb
                               + '_' + level + '_scw_%Y%m%d_v??.cdf')]
        elif instrument == 'fft':
            if level == 'l2':
                pathformat = ('th' + prb + '/' + level + '/' + instrument
                              + '/%Y/th' + prb + '_' + level + '_' + instrument
                              + '_%Y%m%d_v??.cdf')
            elif level == 'l1':
                pathformat = [('th' + prb + '/' + level + '/fff_16/%Y/th' + prb
                               + '_' + level + '_fff_16_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/fff_32/%Y/th' + prb
                               + '_' + level + '_fff_32_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/fff_64/%Y/th' + prb
                               + '_' + level + '_fff_64_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/ffp_16/%Y/th' + prb
                               + '_' + level + '_ffp_16_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/ffp_32/%Y/th' + prb
                               + '_' + level + '_ffp_32_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/ffp_64/%Y/th' + prb
                               + '_' + level + '_ffp_64_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/ffw_16/%Y/th' + prb
                               + '_' + level + '_ffw_16_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/ffw_32/%Y/th' + prb
                               + '_' + level + '_ffw_32_%Y%m%d_v??.cdf'),
                              ('th' + prb + '/' + level + '/ffw_64/%Y/th' + prb
                               + '_' + level + '_ffw_64_%Y%m%d_v??.cdf')]
        elif instrument == 'fbk':
            pathformat = ('th' + prb + '/' + level + '/' + instrument
                          + '/%Y/th' + prb + '_' + level + '_' + instrument
                          + '_%Y%m%d_v??.cdf')
        elif instrument == 'esa':
            pathformat = ('th' + prb + '/' + level + '/' + instrument
                          + '/%Y/th' + prb + '_' + level + '_' + instrument
                          + '_%Y%m%d_v??.cdf')
        elif instrument == 'sst':
            pathformat = ('th' + prb + '/' + level + '/' + instrument
                          + '/%Y/th' + prb + '_' + level + '_' + instrument
                          + '_%Y%m%d_v??.cdf')
        elif instrument == 'mom':
            pathformat = ('th' + prb + '/' + level + '/' + instrument
                          + '/%Y/th' + prb + '_' + level + '_' + instrument
                          + '_%Y%m%d_v??.cdf')
        elif instrument == 'gmom':
            pathformat = ('th' + prb + '/' + level + '/' + instrument
                          + '/%Y/th' + prb + '_' + level + '_' + instrument
                          + '_%Y%m%d_v??.cdf')
        elif instrument == 'state':
            pathformat = ('th' + prb + '/' + level + '/' + instrument
                          + '/%Y/th' + prb + '_' + level + '_' + instrument
                          + '_%Y%m%d.cdf')
        elif instrument == 'slp':
            # note: v01 hard-coded in IDL version as well
            pathformat = 'slp/' + level + '/eph/%Y/slp_l1_eph_%Y%m%d_v01.cdf'
        elif instrument == 'gmag':
            if stations is None:
                logging.error('No stations specified')
                return
            else:
                pathformat = []
                for site, in_greenland in zip(stations, greenland):
                    if in_greenland:
                        pathformat.append('thg/greenland_gmag/' + level
                                          + '/mag/' + site + '/%Y/thg_' + level
                                          + '_mag_' + site + '_%Y%m%d_v??.cdf')
                    else:
                        pathformat.append('thg/' + level + '/mag/' + site
                                          + '/%Y/thg_' + level + '_mag_' + site
                                          + '_%Y%m%d_v??.cdf')

        if not isinstance(pathformat, list):
            pathformat = [pathformat]

        for file_format in pathformat:
            # find the full remote path names using the trange
            remote_names = dailynames(file_format=file_format, trange=trange, res=file_resolution)

            files = download(remote_file=remote_names,
                             remote_path=CONFIG['remote_data_dir'],
                             local_path=CONFIG['local_data_dir'],
                             no_download=no_update,
                             last_version=True)
            if files is not None:
                for file in files:
                    out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files,
                         suffix=suffix,
                         get_support_data=get_support_data,
                         varformat=varformat,
                         varnames=varnames,
                         notplot=notplot)

    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars




import os
import unittest
from pyspedas.utilities.data_exists import data_exists

import pyspedas

class LoadTestCases(unittest.TestCase):
    def test_load_vhm_data(self):
        data = pyspedas.ulysses.vhm()
        self.assertTrue(data_exists('B_MAG'))

    def test_load_swoops_data(self):
        data = pyspedas.ulysses.swoops()
        self.assertTrue(data_exists('Density'))
        self.assertTrue(data_exists('Temperature'))
        self.assertTrue(data_exists('Velocity'))

    def test_load_swics_data(self):
        data = pyspedas.ulysses.swics()
        self.assertTrue(data_exists('Velocity'))

    def test_load_epac_data(self):
        data = pyspedas.ulysses.epac()
        self.assertTrue(data_exists('Omni_Protons'))

    def test_load_hiscale_data(self):
        data = pyspedas.ulysses.hiscale()
        self.assertTrue(data_exists('Electrons'))
        
    def test_load_grb_data(self):
        data = pyspedas.ulysses.grb()
        self.assertTrue(data_exists('Count_Rate'))

    def test_downloadonly(self):
        files = pyspedas.ulysses.urap(downloadonly=True, trange=['2003-01-01', '2003-01-02'])
        self.assertTrue(os.path.exists(files[0]))

if __name__ == '__main__':
    unittest.main()


"""
Perform polarisation analysis of three orthogonal component time series data.

Assumes data are in righthanded fieldaligned coordinate system
with Z pointing the direction of the ambient magnetic field.

The program outputs five spectral results derived from the
fourier transform of the covariance matrix (spectral matrix).

Similar to wavpol.pro in SPEDAS.

Notes:
-----
The program outputs five spectral results derived from the
  fourier transform of the covariance matrix (spectral matrix)

These are follows:

1. Wave power: On a linear scale
    (units of nT^2/Hz if input Bx, By, Bz are in nT)

2. Degree of Polarisation:
    This is similar to a measure of coherency between the input
    signals, however unlike coherency it is invariant under
    coordinate transformation and can detect pure state waves
    which may exist in one channel only.100% indicates a pure
    state wave. Less than 70% indicates noise. For more
    information see J. C. Samson and J. V. Olson 'Some comments
    on the description of the polarization states
    of waves' Geophys. J. R. Astr. Soc. (1980) v61 115-130

3. Wavenormal Angle:
    The angle between the direction of minimum variance
    calculated from the complex off diagonal elements of the
    spectral matrix and the Z direction of the input ac field data.
    for magnetic field data in field aligned coordinates this is the
    wavenormal angle assuming a plane wave. See:
    Means, J. D. (1972), Use of the three-dimensional covariance
    matrix in analyzing the polarization properties of plane waves,
    J. Geophys. Res., 77(28), 5551-5559, doi:10.1029/JA077i028p05551.

4. Ellipticity:
    The ratio (minor axis)/(major axis) of the ellipse transcribed
    by the field variations of the components transverse to the
    Z direction (Samson and Olson, 1980). The sign indicates
    the direction of rotation of the field vector in the plane (cf.
    Means, (1972)).
    Negative signs refer to left-handed rotation about the Z
    direction. In the field aligned coordinate system these signs
    refer to plasma waves of left and right handed polarization.

5. Helicity:
    Similar to Ellipticity except defined in terms of the
    direction of minimum variance instead of Z. Stricltly the Helicity
    is defined in terms of the wavenormal direction or k.
    However since from single point observations the
    sense of k cannot be determined,  helicity here is
    simply the ratio of the minor to major axis transverse to the
    minimum variance direction without sign.

Restrictions:
    If one component is an order of magnitude or more  greater than
    the other two then the polarisation results saturate and erroneously
    indicate high degrees of polarisation at all times and
    frequencies. Time series should be eyeballed before running the program.
    For time series containing very rapid changes or spikes
    the usual problems with Fourier analysis arise.
    Care should be taken in evaluating degree of polarisation results.
    For meaningful results there should be significant wave power at the
    frequency where the polarisation approaches
    100%. Remembercomparing two straight lines yields 100% polarisation.

"""

import warnings
import numpy as np
from pytplot import get_data, store_data, options
from pyspedas import tnames


def atan2c(zx, zy):
    """Define arctan2 with complex numbers."""
    if np.isreal(zx) and np.isreal(zy):
        res = np.arctan2(zx, zy)
    else:
        res = -1j * np.log((zx + 1j*zy)/np.sqrt(zx**2 + zy**2))
    return res


def wpol_ematspec(i1, i2, i3, i4, aa, nosmbins, matspec):
    """Calculate ematspec array."""
    id0 = (i2 - int((nosmbins-1)/2))
    id1 = (i2 + int((nosmbins-1)/2)) + 1
    res = np.sum(aa[0:nosmbins] * matspec[i1, id0:id1, i3, i4])
    return res


def wpol_matsqrd(i1, i2, i3, ematspec):
    """Calculate matsqrd array."""
    res = (ematspec[i1, :, i2, 0] * ematspec[i1, :, 0, i3] +
           ematspec[i1, :, i2, 1] * ematspec[i1, :, 1, i3] +
           ematspec[i1, :, i2, 2] * ematspec[i1, :, 2, i3])
    return res


def wpol_helicity(nosteps, nopfft, KK, ematspec, waveangle):
    """Calculate helicity, ellipticity."""
    # Avoid warnings.
    warnings.simplefilter("ignore", np.ComplexWarning)

    # Define arrays.
    helicity = np.empty((nosteps, int(nopfft/2), 3))
    ellip = np.empty((nosteps, int(nopfft/2), 3))
    alphax = np.empty((nosteps, int(nopfft/2)))
    alphasin2x = np.empty((nosteps, int(nopfft/2)))
    alphacos2x = np.empty((nosteps, int(nopfft/2)))
    alphasin3x = np.empty((nosteps, int(nopfft/2)))
    alphacos3x = np.empty((nosteps, int(nopfft/2)))
    alphay = np.empty((nosteps, int(nopfft/2)))
    alphasin2y = np.empty((nosteps, int(nopfft/2)))
    alphacos2y = np.empty((nosteps, int(nopfft/2)))
    alphasin3y = np.empty((nosteps, int(nopfft/2)))
    alphacos3y = np.empty((nosteps, int(nopfft/2)))
    alphaz = np.empty((nosteps, int(nopfft/2)))
    alphasin2z = np.empty((nosteps, int(nopfft/2)))
    alphacos2z = np.empty((nosteps, int(nopfft/2)))
    alphasin3z = np.empty((nosteps, int(nopfft/2)))
    alphacos3z = np.empty((nosteps, int(nopfft/2)))
    lambdau = np.empty((nosteps, int(nopfft/2), 3, 3), dtype=complex)
    lambday = np.empty((nosteps, int(nopfft/2), 3, 3), dtype=complex)
    lambdaurot = np.empty((nosteps, int(nopfft/2), 2), dtype=complex)

    alphax[KK, :] = np.sqrt(ematspec[KK, :, 0, 0])
    alphacos2x[KK, :] = (np.real(ematspec[KK, :, 0, 1]) /
                         np.sqrt(ematspec[KK, :, 0, 0]))
    alphasin2x[KK, :] = (-np.imag(ematspec[KK, :, 0, 1]) /
                         np.sqrt(ematspec[KK, :, 0, 0]))
    alphacos3x[KK, :] = (np.real(ematspec[KK, :, 0, 2]) /
                         np.sqrt(ematspec[KK, :, 0, 0]))
    alphasin3x[KK, :] = (-np.imag(ematspec[KK, :, 0, 2]) /
                         np.sqrt(ematspec[KK, :, 0, 0]))
    lambdau[KK, :, 0, 0] = alphax[KK, :]
    lambdau[KK, :, 0, 1] = (alphacos2x[KK, :] +
                            1j * alphasin2x[KK, :])
    lambdau[KK, :, 0, 2] = (alphacos3x[KK, :] +
                            1j * alphasin3x[KK, :])

    alphay[KK, :] = np.sqrt(ematspec[KK, :, 1, 1])
    alphacos2y[KK, :] = (np.real(ematspec[KK, :, 1, 0]) /
                         np.sqrt(ematspec[KK, :, 1, 1]))
    alphasin2y[KK, :] = (-np.imag(ematspec[KK, :, 1, 0]) /
                         np.sqrt(ematspec[KK, :, 1, 1]))
    alphacos3y[KK, :] = (np.real(ematspec[KK, :, 1, 2]) /
                         np.sqrt(ematspec[KK, :, 1, 1]))
    alphasin3y[KK, :] = (-np.imag(ematspec[KK, :, 1, 2]) /
                         np.sqrt(ematspec[KK, :, 1, 1]))
    lambdau[KK, :, 1, 0] = alphay[KK, :]
    lambdau[KK, :, 1, 1] = (alphacos2y[KK, :] +
                            1j * alphasin2y[KK, :])
    lambdau[KK, :, 1, 2] = (alphacos3y[KK, :] +
                            1j * alphasin3y[KK, :])

    alphaz[KK, :] = np.sqrt(ematspec[KK, :, 2, 2])
    alphacos2z[KK, :] = (np.real(ematspec[KK, :, 2, 0]) /
                         np.sqrt(ematspec[KK, :, 2, 2]))
    alphasin2z[KK, :] = (-np.imag(ematspec[KK, :, 2, 0]) /
                         np.sqrt(ematspec[KK, :, 2, 2]))
    alphacos3z[KK, :] = (np.real(ematspec[KK, :, 2, 1]) /
                         np.sqrt(ematspec[KK, :, 2, 2]))
    alphasin3z[KK, :] = (-np.imag(ematspec[KK, :, 2, 1]) /
                         np.sqrt(ematspec[KK, :, 2, 2]))
    lambdau[KK, :, 2, 0] = alphaz[KK, :]
    lambdau[KK, :, 2, 1] = (alphacos2z[KK, :] +
                            1j * alphasin2z[KK, :])
    lambdau[KK, :, 2, 2] = (alphacos3z[KK, :] +
                            1j * alphasin3z[KK, :])

    for k in range(int(nopfft/2)):
        for k1 in range(3):
            upper = np.sum(2*np.real(lambdau[KK, k, k1, 0:3]) *
                           np.imag(lambdau[KK, k, k1, 0:3]))
            la2 = np.imag(lambdau[KK, k, k1, 0:3])**2
            lower = np.sum(np.real(lambdau[KK, k, k1, 0:3])**2 - la2)
            gammay = np.nan
            if np.isfinite(upper) and np.isfinite(lower):
                if upper > 0.0:
                    gammay = atan2c(upper, lower)
                else:
                    gammay = (2*np.pi + atan2c(upper, lower))
            lambday[KK, k, k1, :] = (np.exp((0.0 - 1j*0.5*gammay)) *
                                     lambdau[KK, k, k1, :])
            lay2 = np.imag(lambday[KK, k, k1, 0:3])**2
            helicity[KK, k, k1] = (1 /
                                   (np.sqrt(np.real(lambday[KK, k, k1, 0])**2 +
                                    np.real(lambday[KK, k, k1, 1])**2 +
                                    np.real(lambday[KK, k, k1, 2])**2) /
                                    np.sqrt(np.sum(lay2))))
            uppere = (np.imag(lambday[KK, k, k1, 0]) *
                      np.real(lambday[KK, k, k1, 0]) +
                      np.imag(lambday[KK, k, k1, 1]) *
                      np.real(lambday[KK, k, k1, 1]))
            lowere = (-np.imag(lambday[KK, k, k1, 0])**2 +
                      np.real(lambday[KK, k, k1, 0])**2 -
                      np.imag(lambday[KK, k, k1, 1])**2 +
                      np.real(lambday[KK, k, k1, 1])**2)
            gammarot = np.nan
            if np.isfinite(uppere) and np.isfinite(lowere):
                if uppere > 0.0:
                    gammarot = atan2c(uppere, lowere)
                else:
                    gammarot = 2*np.pi + atan2c(uppere, lowere)

            lam = lambday[KK, k, k1, 0:2]
            lambdaurot[KK, k, :] = np.exp(0 - 1j*0.5*gammarot) * lam[:]

            ellip[KK, k, k1] = (np.sqrt(np.imag(lambdaurot[KK, k, 0])**2 +
                                        np.imag(lambdaurot[KK, k, 1])**2) /
                                np.sqrt(np.real(lambdaurot[KK, k, 0])**2 +
                                        np.real(lambdaurot[KK, k, 1])**2))
            ellip[KK, k, k1] = (-ellip[KK, k, k1] *
                                (np.imag(ematspec[KK, k, 0, 1]) *
                                 np.sin(waveangle[KK, k])) /
                                np.abs(np.imag(ematspec[KK, k, 0, 1]) *
                                       np.sin(waveangle[KK, k])))

    # Average over helicity and ellipticity results.
    elliptict0 = (ellip[KK, :, 0]+ellip[KK, :, 1]+ellip[KK, :, 2])/3
    helict0 = (helicity[KK, :, 0]+helicity[KK, :, 1]+helicity[KK, :, 2])/3

    return (helict0, elliptict0)


def wavpol(ct, bx, by, bz,
           nopfft=256,
           steplength=-1,
           bin_freq=3):
    """
    Perform polarisation analysis of Bx, By, Bz time series data.

    Parameters
    ----------
    ct : list of float
        Time.
    b1 : list of float
        Bx field.
    b2 : list of float
        By field.
    b3 : list of float
        Bz field.
    nopfft : int, optional
        Number of points in FFT. The default is 256.
    steplength : int, optional
        The amount of overlap between successive FFT intervals.
        The default is -1 which means nopfft/2.
    bin_freq : int, optional
        Number of bins in frequency domain. The default is 3.

    Returns
    -------
    result: tuple with 9 items

    timeline : list of float
        Times.
    freqline : list of float
        Frequencies.
    powspec : 2-dim array of float
        Wave power.
    degpol : 2-dim array of float
        Degree of Polarisation.
    waveangle : 2-dim array of float
        Wavenormal Angle.
    elliptict : 2-dim array of float
        Ellipticity.
    helict : 2-dim array of float
        Helicity.
    pspec3 : 3-dim array of float
        Power spectra.
    err_flag : bool
        Error flag. The default is 0.
        Returns 1 if there are large number of batches and aborts.

    """
    # Default values.
    if nopfft < 0:
        nopfft = 256
    if steplength < 0:
        steplength = nopfft / 2
    if bin_freq < 0:
        bin_freq = 3

    # Convert to numpy arrays.
    ct = np.array(ct, np.float64)
    bx = np.array(bx, np.float64)
    by = np.array(by, np.float64)
    bz = np.array(bz, np.float64)

    # Define empty returns.
    timeline = ''
    freqline = ''
    powspec = ''
    degpol = ''
    waveangle = ''
    elliptict = ''
    helict = ''
    pspec3 = ''
    err_flag = 0

    # Define variables.
    nopoints = len(bx)
    iano = np.zeros(nopoints, dtype=int)
    dt = [ct[i+1]-ct[i] for i in range(len(ct)-1)]  # time difference
    beginsampfreq = 1./(ct[1]-ct[0])
    endsampfreq = 1./(ct[nopoints-1]-ct[nopoints-2])

    if beginsampfreq != endsampfreq:
        print('wavpol Warning: file sampling frequency changes from ',
              beginsampfreq, 'Hz to ', endsampfreq, 'Hz')
    else:
        print('wavpol: File sampling frequency=', beginsampfreq, 'Hz')

    samp_freq = beginsampfreq
    samp_per = 1./samp_freq

    # Time reversal detection.
    for i in range(len(dt)-1):
        if dt[i] < 0:
            iano[i] = 16

    # The accuracy of the sampling frequency should be about 1%
    accuracy = 0.01

    # Find discontinuities.
    discont_trigger = accuracy*samp_per
    for i in range(nopoints-1):
        if abs(dt[i]-1./samp_freq) > discont_trigger:
            iano[i] = 17
    iano[nopoints-1] = 22

    # Count batches, should be less than 80,000
    n_batches = 0
    errs = []
    for i in range(nopoints):
        if iano[i] >= 15:
            n_batches += 1
            errs.append(i)

    # If there are too many batches, return.
    if n_batches > 80000.0:
        print("wavpol error: Large number of batches. " +
              "Returning to avoid memory runaway.")
        err_flag = 1
        result = (timeline, freqline, powspec, degpol, waveangle,
                  elliptict, helict, pspec3, err_flag)
        return result

    nbp_fft_batches = [0] * n_batches
    # Total numbers of FFT calculations including 1 leap frog for each batch
    ind_batch0 = 0
    nosteps = 0
    print('n_batches', n_batches)

    for i in range(n_batches):
        nosteps = int(nosteps + np.floor((errs[i] - ind_batch0)/steplength))
        ind_batch0 = errs[i]

    nosteps = nosteps + n_batches
    print('Total number of steps:', nosteps)

    # leveltplot = 0.000001  # Power rejection level 0 to 1
    nosmbins = bin_freq  # No. of bins in frequency domain
    # Smoothing profile based on Hanning:
    aa = np.array([0.024, 0.093, 0.232, 0.301, 0.232, 0.093, 0.024])

    ind0 = 0
    KK = 0

    # Create empty arrays
    specx = np.empty((nosteps, nopfft), dtype=complex)
    specy = np.empty((nosteps, nopfft), dtype=complex)
    specz = np.empty((nosteps, nopfft), dtype=complex)
    halfspecx = np.empty((nosteps, int(nopfft/2)), dtype=complex)
    halfspecy = np.empty((nosteps, int(nopfft/2)), dtype=complex)
    halfspecz = np.empty((nosteps, int(nopfft/2)), dtype=complex)
    matspec = np.empty((nosteps, int(nopfft/2), 3, 3), dtype=complex)
    ematspec = np.empty((nosteps, int(nopfft/2), 3, 3), dtype=complex)
    aaa2 = np.empty((nosteps, int(nopfft/2)))
    wnx = np.empty((nosteps, int(nopfft/2)))
    wny = np.empty((nosteps, int(nopfft/2)))
    wnz = np.empty((nosteps, int(nopfft/2)))
    waveangle = np.empty((nosteps, int(nopfft/2)))
    matsqrd = np.empty((nosteps, int(nopfft/2), 3, 3), dtype=complex)
    trmatsqrd = np.empty((nosteps, int(nopfft/2)))
    trmatspec = np.empty((nosteps, int(nopfft/2)))
    degpol = np.empty((nosteps, int(nopfft/2)))
    xrmatspec = np.empty((nosteps, int(nopfft/2)))
    yrmatspec = np.empty((nosteps, int(nopfft/2)))
    zrmatspec = np.empty((nosteps, int(nopfft/2)))

    # Return arrays.
    timeline = np.empty((nosteps))
    freqline = ''
    helict = np.empty((nosteps, int(nopfft/2)))
    elliptict = np.empty((nosteps, int(nopfft/2)))
    powspec = np.empty((nosteps, int(nopfft/2)))
    pspecx = np.empty((nosteps, int(nopfft/2)))
    pspecy = np.empty((nosteps, int(nopfft/2)))
    pspecz = np.empty((nosteps, int(nopfft/2)))
    pspec3 = np.empty((nosteps, int(nopfft/2), 3))

    for batch in range(n_batches):
        ind1 = errs[batch]+1
        # nbp_batch = ind1-ind0+1
        ind1_ref = ind1
        KK_batch_start = KK

        xs = np.array(bx[ind0:ind1])
        ys = np.array(by[ind0:ind1])
        zs = np.array(bz[ind0:ind1])

        ngood = np.count_nonzero(~np.isnan(xs))  # Count finite data.
        if ngood > nopfft:
            nbp_fft_batches[batch] = np.floor(ngood/steplength)
            print('Total number of possible FFT in the batch no ', batch,
                  ' is:', nbp_fft_batches[batch])
            ind0_fft = 0
            for j in range(int(nbp_fft_batches[batch])):
                # ind1_fft = nopfft * (j+1)-1
                # ind1_ref_fft = ind1_fft

                # FFT Calculation.
                smooth = 0.08 + 0.46 * (1 - np.cos(2 * np.pi *
                                                   np.arange(nopfft) / nopfft))
                tempx = smooth * xs[0:nopfft]
                tempy = smooth * ys[0:nopfft]
                tempz = smooth * zs[0:nopfft]

                # previous version (prior to 15Dec2021)
                # specx[KK, :] = np.fft.fft(tempx, norm='forward')
                # specy[KK, :] = np.fft.fft(tempy, norm='forward')
                # specz[KK, :] = np.fft.fft(tempz, norm='forward')
                specx[KK, :] = np.fft.fft(tempx, norm='ortho')/len(tempx)
                specy[KK, :] = np.fft.fft(tempy, norm='ortho')/len(tempy)
                specz[KK, :] = np.fft.fft(tempz, norm='ortho')/len(tempz)

                halfspecx[KK, :] = specx[KK, 0:int(nopfft/2)]
                halfspecy[KK, :] = specy[KK, 0:int(nopfft/2)]
                halfspecz[KK, :] = specz[KK, 0:int(nopfft/2)]
                xs = np.roll(xs, -int(steplength))
                ys = np.roll(ys, -int(steplength))
                zs = np.roll(zs, -int(steplength))

                # Calculation of the spectral matrix.
                matspec[KK, :, 0, 0] = (halfspecx[KK, :] *
                                        np.conjugate(halfspecx[KK, :]))
                matspec[KK, :, 1, 0] = (halfspecx[KK, :] *
                                        np.conjugate(halfspecy[KK, :]))
                matspec[KK, :, 2, 0] = (halfspecx[KK, :] *
                                        np.conjugate(halfspecz[KK, :]))
                matspec[KK, :, 0, 1] = (halfspecy[KK, :] *
                                        np.conjugate(halfspecx[KK, :]))
                matspec[KK, :, 1, 1] = (halfspecy[KK, :] *
                                        np.conjugate(halfspecy[KK, :]))
                matspec[KK, :, 2, 1] = (halfspecy[KK, :] *
                                        np.conjugate(halfspecz[KK, :]))
                matspec[KK, :, 0, 2] = (halfspecz[KK, :] *
                                        np.conjugate(halfspecx[KK, :]))
                matspec[KK, :, 1, 2] = (halfspecz[KK, :] *
                                        np.conjugate(halfspecy[KK, :]))
                matspec[KK, :, 2, 2] = (halfspecz[KK, :] *
                                        np.conjugate(halfspecz[KK, :]))

                # Calculation of smoothed spectral matrix.
                for k in range(int((nosmbins-1)/2),
                               int((nopfft/2-1)-(nosmbins-1)/2)+1):
                    for k2 in range(0, 3):
                        for k1 in range(0, 3):
                            ematspec[KK, k, k1, k2] = wpol_ematspec(KK, k, k1,
                                                                    k2, aa,
                                                                    nosmbins,
                                                                    matspec)

                # Calculation of the minimum variance direction
                # and wavenormal angle.
                aaa2[KK, :] = np.sqrt(np.imag(ematspec[KK, :, 0, 1])**2 +
                                      np.imag(ematspec[KK, :, 0, 2])**2 +
                                      np.imag(ematspec[KK, :, 1, 2])**2)
                # Avoid warnings due to NaN values.
                np.seterr(divide='ignore', invalid='ignore')
                wnx[KK, :] = np.abs(np.imag(ematspec[KK, :, 1, 2]) /
                                    aaa2[KK, :])
                wny[KK, :] = -np.abs(np.imag(ematspec[KK, :, 0, 2]) /
                                     aaa2[KK, :])
                wnz[KK, :] = (np.imag(ematspec[KK, :, 0, 1]) / aaa2[KK, :])
                waveangle[KK, :] = np.arctan2(np.sqrt(wnx[KK, :]**2 +
                                                      wny[KK, :]**2),
                                              np.abs(wnz[KK, :]))

                # Calculation of the degree of polarization.
                # Calculation of square of smoothed spec matrix.
                for k1 in range(3):
                    for k2 in range(3):
                        matsqrd[KK, :, k1, k2] = wpol_matsqrd(KK, k1, k2,
                                                              ematspec)

                trmatsqrd[KK, :] = np.real(matsqrd[KK, :, 0, 0] +
                                           matsqrd[KK, :, 1, 1] +
                                           matsqrd[KK, :, 2, 2])
                trmatspec[KK, :] = np.real(ematspec[KK, :, 0, 0] +
                                           ematspec[KK, :, 1, 1] +
                                           ematspec[KK, :, 2, 2])
                id1 = int((nosmbins-1)/2)
                id2 = int((nopfft/2-1)-(nosmbins-1)/2) + 1
                degpol[KK, id1:id2] = ((3 * trmatsqrd[KK, id1:id2] -
                                        trmatspec[KK, id1:id2]**2) /
                                       (2 * trmatspec[KK, id1:id2]**2))

                xrmatspec[KK, :] = np.real(ematspec[KK, :, 0, 0])
                yrmatspec[KK, :] = np.real(ematspec[KK, :, 1, 1])
                zrmatspec[KK, :] = np.real(ematspec[KK, :, 2, 2])

                # Calculation of helicity, ellipticity
                # and the wave state vector
                (helict[KK], elliptict[KK]) = wpol_helicity(nosteps,
                                                            nopfft,
                                                            KK,
                                                            ematspec,
                                                            waveangle)

                # Scaling power results to units with meaning
                binwidth = samp_freq / nopfft
                W = np.sum(smooth**2) / np.real(nopfft)
                id2 = int(nopfft/2-1)
                powspec[KK, 1:id2] = 1./W*2*trmatspec[KK, 1:id2]/binwidth
                powspec[KK, 0] = 1./W * trmatspec[KK, 0]/binwidth
                powspec[KK, id2] = 1./W*trmatspec[KK, id2]/binwidth

                pspecx[KK, 1:id2] = 1./W*2*xrmatspec[KK, 1:id2]/binwidth
                pspecx[KK, 0] = 1./W*xrmatspec[KK, 0]/binwidth
                pspecx[KK, id2] = 1./W*xrmatspec[KK, id2]/binwidth

                pspecy[KK, 1:id2] = 1./W*2*yrmatspec[KK, 1:id2]/binwidth
                pspecy[KK, 0] = 1./W*yrmatspec[KK, 0]/binwidth
                pspecy[KK, id2] = 1./W*yrmatspec[KK, id2]/binwidth

                pspecz[KK, 1:id2] = 1./W*2*zrmatspec[KK, 1:id2]/binwidth
                pspecz[KK, 0] = 1./W*zrmatspec[KK, 0]/binwidth
                pspecz[KK, id2] = 1./W*zrmatspec[KK, id2]/binwidth

                pspec3[KK, :, 0] = pspecx[KK, :]
                pspec3[KK, :, 1] = pspecy[KK, :]
                pspec3[KK, :, 2] = pspecz[KK, :]

                ind0_fft = ind0_fft + steplength
                KK_batch_stop = KK

                # Print an indication that a computation is happening.
                if KK == 0 or KK % 40 == 0:
                    print(' ')
                    print('wavpol step', KK, ' ', end='')
                elif KK % 4 == 0:
                    print('.', end='')

                KK += 1
                # End loop "for j"

            ids = KK_batch_start
            idf = KK_batch_stop+1
            ta = np.arange(nbp_fft_batches[batch])
            timeline[ids:idf] = (ct[ind0] +
                                 np.abs(int(nopfft/2))/samp_freq +
                                 ta*steplength/samp_freq)
            if KK == len(timeline):
                continue

            timeline[idf] = (ct[ind0] +
                             np.abs(int(nopfft/2))/samp_freq +
                             (nbp_fft_batches[batch]+1)*steplength/samp_freq)
            KK += 1
            # End "if ngood > nopfft"
        else:
            binwidth = samp_freq/nopfft
            print('Fourier Transform is not possible. ',
                  'Ngood = ', ngood,
                  'Required number of points for FFT = ', nopfft)

            timeline[KK] = (ct[ind0] +
                            np.abs(int(nopfft/2))/samp_freq +
                            steplength/samp_freq)
            powspec[KK, 0:int(nopfft/2)+1] = np.nan
            KK += 1
        ind0 = ind1_ref + 1
        # End "for batch in range(n_batches)"

    freqline = binwidth*np.arange(int(nopfft/2))

    # Make sure there aren't any missing data points at the end of the output.
    wherezero = np.argwhere(timeline == 0)
    if len(wherezero) > 0:
        timeline[wherezero] = np.nan
        powspec[wherezero, :] = np.nan
        pspecx[wherezero, :] = np.nan
        pspecy[wherezero, :] = np.nan
        pspecz[wherezero, :] = np.nan
        pspec3[wherezero, :, :] = np.nan
        elliptict[wherezero, :] = np.nan
        helict[wherezero, :] = np.nan

    # Returns results.
    result = (timeline, freqline, powspec, degpol, waveangle,
              elliptict, helict, pspec3, err_flag)
    print('\nwavpol completed successfully')

    return result


def twavpol(tvarname, prefix='', nopfft=-1, steplength=-1, bin_freq=-1):
    """Apply wavpol to a pytplot variable.

    Creates multiple pytplot variables:
    '_powspec','_degpol', '_waveangle', '_elliptict', '_helict',
    '_pspec3_x', '_pspec3_y', '_pspec3_z'

    Parameters
    ----------
    tvarname : string
              Name of pytplot variable.
    prefix : string, optional
            Prefix for pytplot variables created.
    nopfft : int, optional
        Number of points in FFT. The default is 256.
    steplength : int, optional
        The amount of overlap between successive FFT intervals.
        The default is -1 which means nopfft/2.
    bin_freq : int, optional
        Number of bins in frequency domain. The default is 3.

    Returns
    -------
    result : bool
        Returns 1 if completed successfully.
        Returns 0 if it encountered problems and exited.
    """
    if prefix == '':
        prefix = tvarname

    all_names = tnames(tvarname)
    if len(all_names) < 1:
        print('twavpol error: No valid pytplot variables match tvarname.')
        return 0

    xdata = get_data(tvarname)

    ct = xdata.times
    if len(ct) < 2:
        print('twavpol error: Time variable does not have enough points.')
        return 0

    bfield = xdata.y
    if bfield.ndim != 2:
        print('twavpol error: Data should have 2 dimensions.')
        return 0
    b1 = bfield[:, 0]
    b2 = bfield[:, 1]
    b3 = bfield[:, 2]
    if (len(ct) != len(b1) or len(ct) != len(b2) or len(ct) != len(b3)):
        print('twavpol error: Number of time elements does not match' +
              'number of magnetic field elements.')
        return 0

    # Apply vawpol.
    (timeline, freqline, powspec, degpol, waveangle, elliptict,
     helict, pspec3, err_flag) = wavpol(ct, b1, b2, b3,
                                        nopfft=nopfft, steplength=steplength,
                                        bin_freq=bin_freq)

    if err_flag == 1:
        print('twavpol error: There were errors while applying wavpol.')
        return 0

    # Store new pytplot variables as spectrograms.
    vt = prefix+'_powspec'
    store_data(vt, data={'x': timeline, 'y': powspec, 'v': freqline})
    options(vt, 'spec', 1)
    vt = prefix+'_degpol'
    store_data(vt, data={'x': timeline, 'y': degpol, 'v': freqline})
    options(vt, 'spec', 1)
    vt = prefix+'_waveangle'
    store_data(vt, data={'x': timeline, 'y': waveangle, 'v': freqline})
    options(vt, 'spec', 1)
    vt = prefix+'_elliptict'
    store_data(vt, data={'x': timeline, 'y': elliptict, 'v': freqline})
    options(vt, 'spec', 1)
    vt = prefix+'_helict'
    store_data(vt, data={'x': timeline, 'y': helict, 'v': freqline})
    options(vt, 'spec', 1)

    # Take the three components of pspec3.
    vt = prefix+'_pspec3_x'
    store_data(vt, data={'x': timeline, 'y': pspec3[:, :, 0], 'v': freqline})
    options(vt, 'spec', 1)
    vt = prefix+'_pspec3_y'
    store_data(vt, data={'x': timeline, 'y': pspec3[:, :, 1], 'v': freqline})
    options(vt, 'spec', 1)
    vt = prefix+'_pspec3_z'
    store_data(vt, data={'x': timeline, 'y': pspec3[:, :, 2], 'v': freqline})
    options(vt, 'spec', 1)

    return 1



from pyspedas import time_string, time_double
from pyspedas.analysis.time_clip import time_clip as tclip
from pyspedas.hapi.hapi import hapi
from .config import CONFIG

def load(trange=['2017-03-27/06:00', '2017-03-27/08:00'], 
         probe='a',
         instrument='mag',
         datatype='', 
         level='l1b', 
         suffix='', 
         varnames=[], # should be sent to hapi as parameters?
         time_clip=False):
    """
    This function loads data from the Swarm mission; this function is not meant 
    to be called directly; instead, see the wrappers:
        pyspedas.swarm.mag

    """
    server = CONFIG['remote_data_dir'] + 'hapi/'

    if not isinstance(probe, list):
        probe = [probe]

    tr = time_string(time_double(trange), fmt='%Y-%m-%dT%H:%M:%S')

    out_vars = []

    for this_probe in probe:
        prefix = 'swarm' + this_probe.lower() + '_'
        if instrument == 'mag':
            dataset = 'SW_OPER_MAG' + this_probe.upper() + '_' + datatype.upper() + '_' + level[1:].upper()
            tvars = hapi(trange=tr, 
                         server=server, 
                         dataset=dataset, 
                         parameters=varnames, 
                         suffix=suffix, 
                         prefix=prefix)

            if tvars is not None and len(tvars) != 0:
                out_vars.extend(tvars)

    if time_clip:
        for new_var in out_vars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return out_vars




from pyspedas.particles.moments.moments_3d import moments_3d

def spd_pgs_moments(data_in, sc_pot=0):
    """

    Calculates moments from a simplified particle data structure.
    Simply a wrapper for moments_3d right now

    Input:
        data_in: dict
            Particle data structure 

    Parameters:
        sc_pot: float
            Spacecraft potential

    Returns:
        Dictionary containing moments
    """

    return moments_3d(data_in, sc_pot=sc_pot)



import logging
from pytplot import store_data, options

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)

def spd_pgs_moments_tplot(moments, x=None, prefix='', suffix=''):
    """
    Creates tplot variables from moments dictionaries

    Input:
        moments: dict
            Dictionary containing moments values returned by moments_3d

    Parameters:
        x: numpy.ndarray
            The x-axis (time) values

        prefix: str
            Name prefix for the output variables

        suffix: str
            Name suffix for the output variables

    Returns:
        List of tplot variables created.
    """

    if x is None:
        logging.error('Error, no x-values specified')
        return

    if not isinstance(moments, dict):
        logging.error('Error, the "moments" variable must be a hash table containing the moments')
        return


    for key in moments.keys():
        store_data(prefix + '_' + key + suffix, data={'x': x, 'y': moments[key]})

    options(prefix + '_density' + suffix, 'ysubtitle', '[1/cc]')
    options(prefix + '_velocity' + suffix, 'yrange', [-800, 800])
    options(prefix + '_flux' + suffix, 'yrange', [-1e8, 1e8])

    return [prefix + '_' + key + suffix for key in moments.keys()]



"""
Interpolates data specified in 'names' to time values found in 'interp_to'.

Notes
-----
Allowed wildcards are ? for a single character, * from multiple characters.
Similar to tinterpol.pro in IDL SPEDAS.

"""

from pyspedas import tnames
from pytplot import get_data, store_data

def tinterpol(names, interp_to, method=None, newname=None, suffix=None):
    """
    Interpolate data to times in interp_to.

    Parameters
    ----------
    names: str/list of str
        List of variables to interpolate.
    interp_to: str
        String containing the variable
            containing the time stamps to interpolate to
    method: str, optional
        Interpolation method. Default is ‘linear’.
        Specifies the kind of interpolation as a string (‘linear’, ‘nearest’,
        ‘zero’, ‘slinear’, ‘quadratic’, ‘cubic’, ‘previous’, ‘next’)
        where ‘zero’, ‘slinear’, ‘quadratic’ and ‘cubic’ refer to a spline
        interpolation of zeroth, first, second or third order; ‘previous’ and
        ‘next’ simply return the previous or next value of the point) or
        as an integer specifying the order of the spline interpolator to use.
    newname: str/list of str, optional
        List of new_names for pytplot variables.
        If '', then pytplot variables are replaced.
        If not given, then a suffix is applied.
    suffix: str, optional
        A suffix to apply. Default is '-itrp'.

    Returns
    -------
    None.

    """
    if not isinstance(names, list):
        names = [names]
    if not isinstance(newname, list):
        newname = [newname]

    old_names = tnames(names)

    if len(old_names) < 1:
        print('tinterpol error: No pytplot names were provided.')
        return

    if suffix is None:
        suffix = '-itrp'

    if method is None:
        method = 'linear'

    if (newname is None) or (len(newname) == 1 and newname[0] is None):
        n_names = [s + suffix for s in old_names]
    elif newname == '':
        n_names = old_names
    else:
        n_names = newname

    if isinstance(interp_to, str):
        interp_to_data = get_data(interp_to)

        if interp_to_data is None:
            print('Error, tplot variable: ' + interp_to + ' not found.')
            return

        interp_to_times = interp_to_data[0]
    else:
        interp_to_times = interp_to

    for name_idx, name in enumerate(old_names):
        xdata = get_data(name, xarray=True)
        xdata_interpolated = xdata.interp({'time': interp_to_times},
                                          method=method)

        if 'spec_bins' in xdata.coords:
            store_data(n_names[name_idx],
                       data={
                        'x': interp_to_times,
                        'y': xdata_interpolated.values,
                        'v': xdata_interpolated.coords['spec_bins'].values
                        })
        else:
            store_data(n_names[name_idx], data={'x': interp_to_times,
                       'y': xdata_interpolated.values})

        print('tinterpol (' + method + ') was applied to: '
              + n_names[name_idx])




import re
from dateutil.parser import parse
from bisect import bisect_left

def mms_files_in_interval(in_files, trange):
    '''
    This function filters the file list returned by the SDC to the requested time range. This filter is purposefully 
    liberal, it regularly grabs an extra file due to special cases

    Parameters:
        in_files: list of dict
            List of hash tables containing files returned by the SDC

        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

    Returns:
        List of hash tables containing file names, sizes and their time tags

    '''
    file_name = 'mms.*_([0-9]{8,14})_v(\d+).(\d+).(\d+).cdf'

    file_times = []

    regex = re.compile(file_name)

    for file in in_files:
        matches = regex.match(file['file_name'])
        if matches:
            file_times.append((file['file_name'], parse(matches.groups()[0]).timestamp(), file['timetag'], file['file_size']))

    # sort in time
    sorted_files = sorted(file_times, key=lambda x: x[1])

    times = [t[1] for t in sorted_files]

    idx_min = bisect_left(times, parse(trange[0]).timestamp())

    # note: purposefully liberal here; include one extra file so that we always get the burst mode data
    if idx_min == 0:
        return [{'file_name': f[0], 'timetag': f[2], 'file_size': f[3]} for f in sorted_files[idx_min:]]
    else:
        return [{'file_name': f[0], 'timetag': f[2], 'file_size': f[3]} for f in sorted_files[idx_min-1:]]




import numpy as np

from pyspedas.particles.moments.moments_3d_omega_weights import moments_3d_omega_weights

# use nansum from bottleneck if it's installed, otherwise use the numpy one
try:
    import bottleneck as bn
    nansum = bn.nansum
except ImportError:
    nansum = np.nansum

def moments_3d(data_in, sc_pot=0, no_unit_conversion=False):
    """
    Calculates plasma moments from 3D data structure

    Input:
        data_in: dict
            Particle data structure 

    Parameters:
        sc_pot: float
            Spacecraft potential

        no_unit_conversion: bool
            Flag indicating that datta is already in eflux and no unit
            conversion is required

    Notes:
        The calculations were mostly heisted from Davin Larson's IDL version

    Returns:
        Dictionary containing moments

    """

    charge = data_in['charge']
    mass = data_in['mass']
    energy = data_in['energy']
    energy[energy < 0.1] = 0.1

    de = data_in['denergy']
    de_e = de/energy

    e_inf = energy + charge*sc_pot
    e_inf[e_inf < 0] = 0.0

    # mystery line from the IDL version
    weight = (energy + charge*sc_pot)/de + 0.5
    weight[weight < 0] = 0
    weight[weight > 1] = 1

    domega_weight = moments_3d_omega_weights(data_in['theta'], data_in['phi'], data_in['dtheta'], data_in['dphi'])

    data_dv = data_in['data']*de_e*weight*domega_weight[0, :, :]

    # density calculation
    dweight = np.sqrt(e_inf)/energy
    pardens = np.sqrt(mass/2.0)*1e-5*data_dv*dweight
    density = nansum(pardens)

    # flux calculation
    tmp = data_in['data']*de_e*weight*e_inf/energy
    fx = nansum(tmp*domega_weight[1, :, :])
    fy = nansum(tmp*domega_weight[2, :, :])
    fz = nansum(tmp*domega_weight[3, :, :])

    flux = np.array([fx, fy, fz])

    # velocity flux calculation
    tmp = data_in['data']*de_e*weight*e_inf**1.5/energy
    vfxx = nansum(tmp*domega_weight[4, :, :])
    vfyy = nansum(tmp*domega_weight[5, :, :])
    vfzz = nansum(tmp*domega_weight[6, :, :])
    vfxy = nansum(tmp*domega_weight[7, :, :])
    vfxz = nansum(tmp*domega_weight[8, :, :])
    vfyz = nansum(tmp*domega_weight[9, :, :])

    vftens = np.array([vfxx, vfyy, vfzz, vfxy, vfxz, vfyz])*np.sqrt(2.0/mass)*1e5
    mftens = vftens*mass/1e10

    # energy flux calculation (extra factor of energy)
    tmp = data_in['data']*de_e*weight*e_inf**2/energy
    v2f_x = nansum(tmp*domega_weight[1, :, :])
    v2f_y = nansum(tmp*domega_weight[2, :, :])
    v2f_z = nansum(tmp*domega_weight[3, :, :])
    eflux = np.array([v2f_x, v2f_y, v2f_z])

    velocity = flux/density/1e5 # km/s

    mf3x3 = np.array([[mftens[0], mftens[3], mftens[4]], [mftens[3], mftens[1], mftens[5]], [mftens[4], mftens[5], mftens[2]]])
    pt3x3 = mf3x3 - np.outer(velocity, flux)*mass/1e5
    ptens = np.array([pt3x3[0, 0], pt3x3[1, 1], pt3x3[2, 2], pt3x3[0, 1], pt3x3[0, 2], pt3x3[1, 2]])

    t3x3 = pt3x3/density
    avgtemp = (t3x3[0, 0] + t3x3[1, 1] + t3x3[2, 2] )/3.0  # trace/3

    vthermal = np.sqrt(2.0*avgtemp/mass)

    # t3 = sp.linalg.svdvals(t3x3)
    
    output = {'density': density, 
              'flux': flux, 
              'mftens': mftens, 
              'velocity': velocity, 
              'ptens': ptens, 
              'ttens': t3x3, 
              'vthermal': vthermal,
              'avgtemp': avgtemp}
    return output



import numpy as np
from pytplot import get_data, store_data
from geopack import geopack, t04

def tts04(pos_var_gsm, parmod=None, suffix=''):
    """
    tplot wrapper for the functional interface to Sheng Tian's implementation of the 
    Tsyganenko-Sitnov (2004) storm-time geomagnetic field model

    https://github.com/tsssss/geopack

    Input
    ------
        pos_gsm_tvar: str
            tplot variable containing the position data (km) in GSM coordinates

    Parameters
    -----------
        parmod: ndarray
            10-element array (vs. time):
                (1) solar wind pressure pdyn (nanopascals),
                (2) dst (nanotesla),
                (3) byimf,
                (4) bzimf (nanotesla)
                (5-10) indices w1 - w6, calculated as time integrals from the beginning of a storm
                    see the reference (3) below, for a detailed definition of those variables

        suffix: str
            Suffix to append to the tplot output variable

    Returns
    --------
        Name of the tplot variable containing the model data
    """
    pos_data = get_data(pos_var_gsm)

    if pos_data is None:
        print('Variable not found: ' + pos_var_gsm)
        return

    b0gsm = np.zeros((len(pos_data.times), 3))
    dbgsm = np.zeros((len(pos_data.times), 3))

    # convert to Re
    pos_re = pos_data.y/6371.2

    if parmod is not None:
        par = get_data(parmod)

        if par is not None:
            par = par.y
    else:
        print('parmod keyword required.')
        return

    for idx, time in enumerate(pos_data.times):
        tilt = geopack.recalc(time)

        if not np.isfinite(par[idx, :]).all():
            # skip if there are any NaNs in the input
            continue

        # dipole B in GSM
        b0gsm[idx, 0], b0gsm[idx, 1], b0gsm[idx, 2] = geopack.dip(pos_re[idx, 0], pos_re[idx, 1], pos_re[idx, 2])

        # T96 dB in GSM
        dbgsm[idx, 0], dbgsm[idx, 1], dbgsm[idx, 2] = t04.t04(par[idx, :], tilt, pos_re[idx, 0], pos_re[idx, 1], pos_re[idx, 2])

    bgsm = b0gsm + dbgsm

    saved = store_data(pos_var_gsm + '_bts04' + suffix, data={'x': pos_data.times, 'y': bgsm})

    if saved:
        return pos_var_gsm + '_bts04' + suffix



import math
import numpy as np
from pytplot import get_data, store_data, options
from pyspedas import tinterpol

def mms_curl(fields=None, positions=None, suffix=''):
    """
    This function applies the curlometer technique to MMS FGM data
    
    Parameters
    ----------
        fields : list of str
            List of tplot variables containing the B-field for each spacecraft 
            (in GSE coordinates)

        positions : list of str
            List of tplot variables containing the S/C position vectors for 
            each spacecraft (also GSE coordinates) 

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

    Notes
    ----------
        The input B-field data and position data are required to be in 
        GSE coordinates
 
        Based on the original mms_curl, written in IDL, by Jonathan Eastwood 

        For more info on this method, see:
          Chanteur, G., Spatial Interpolation for Four Spacecraft: Theory, 
          Chapter 14 of Analysis methods for multi-spacecraft data, G. 
          Paschmann and P. W. Daly (Eds.) ISSI Scientific Report SR-001. 

    Returns
    ----------
        List of tplot variables created

    """
    if fields is None or positions is None:
        print('Error: B-field and spacecraft position keywords required.')
        return

    if len(fields) != 4 or len(positions) != 4:
        print('Error, fields and positions keywords should be specified as 4-element arrays containing the tplot variable name for the field and position variables')
        return
        
    # *********************************************************
    # Magnetic Field
    # *********************************************************
    # interpolate the magnetic field data all onto the same timeline (MMS1):
    # should be in GSE coordinates
    tinterpol(fields[1], fields[0], newname=fields[1] + '_i')
    tinterpol(fields[2], fields[0], newname=fields[2] + '_i')
    tinterpol(fields[3], fields[0], newname=fields[3] + '_i')

    # interpolate the definitive ephemeris onto the magnetic field timeseries
    # should be in GSE coordinates
    tinterpol(positions[0], fields[0], newname=positions[0] + '_i')
    tinterpol(positions[1], fields[0], newname=positions[1] + '_i')
    tinterpol(positions[2], fields[0], newname=positions[2] + '_i')
    tinterpol(positions[3], fields[0], newname=positions[3] + '_i')

    m0 = 4.0*math.pi*1e-7

    mms1_bfield = get_data(fields[0])
    mms2_bfield = get_data(fields[1] + '_i')
    mms3_bfield = get_data(fields[2] + '_i')
    mms4_bfield = get_data(fields[3] + '_i')

    if mms1_bfield is None:
        print('Error, B-field variable is missing: ' + fields[0])
        return
    elif mms2_bfield is None:
        print('Error, B-field variable is missing: ' + fields[1] + '_i')
        return
    elif mms3_bfield is None:
        print('Error, B-field variable is missing: ' + fields[2] + '_i')
        return
    elif mms4_bfield is None:
        print('Error, B-field variable is missing: ' + fields[3] + '_i')
        return

    timesb1, datab1 = mms1_bfield
    timesb2, datab2 = mms2_bfield
    timesb3, datab3 = mms3_bfield
    timesb4, datab4 = mms4_bfield

    # extract the vector
    b1 = datab1[:, 0:3]
    b2 = datab2[:, 0:3]
    b3 = datab3[:, 0:3]
    b4 = datab4[:, 0:3]

    mms1_pos = get_data(positions[0] + '_i')
    mms2_pos = get_data(positions[1] + '_i')
    mms3_pos = get_data(positions[2] + '_i')
    mms4_pos = get_data(positions[3] + '_i')

    if mms1_pos is None:
        print('Error, S/C position variable is missing: ' + positions[0] + '_i')
        return
    elif mms2_pos is None:
        print('Error, S/C position variable is missing: ' + positions[1] + '_i')
        return
    elif mms3_pos is None:
        print('Error, S/C position variable is missing: ' + positions[2] + '_i')
        return
    elif mms4_pos is None:
        print('Error, S/C position variable is missing: ' + positions[3] + '_i')
        return

    timesp1, p1 = mms1_pos
    timesp2, p2 = mms2_pos
    timesp3, p3 = mms3_pos
    timesp4, p4 = mms4_pos

    divb = np.zeros([len(timesb1), 5])
    baryb = np.zeros([len(timesb1), 3])
    baryb2 = np.zeros([len(timesb1), 3])
    baryb3 = np.zeros([len(timesb1), 3])
    baryb4 = np.zeros([len(timesb1), 3])

    jtotal = np.zeros([len(timesb1), 4])
    btotal = np.zeros([len(timesb1), 1])
    jparallel = np.zeros([len(timesb1), 1])
    jperpvec = np.zeros([len(timesb1), 4])
    alphaparallel = np.zeros([len(timesb1), 1])
    alpha = np.zeros([len(timesb1), 1])

    # leave as a loop for now because you have to construct and manipulate a matrix for each time step.
    for i, time in enumerate(timesb1):
        p12 = p2[i, 0:3]-p1[i, 0:3]
        p13 = p3[i, 0:3]-p1[i, 0:3]
        p14 = p4[i, 0:3]-p1[i, 0:3]

        k2 = np.cross(p13, p14)*(1/(np.matmul(p12, np.transpose(np.cross(p13, p14)))))
        k3 = np.cross(p12, p14)*(1/(np.matmul(p13, np.transpose(np.cross(p12, p14)))))
        k4 = np.cross(p12, p13)*(1/(np.matmul(p14, np.transpose(np.cross(p12, p13)))))

        k1 = 0-k4-k3-k2

        curlmag = np.cross(k1, b1[i, :])+np.cross(k2, b2[i, :])+np.cross(k3, b3[i, :])+np.cross(k4, b4[i, :])
        divergence = np.matmul(b1[i, :], k1) + np.matmul(b2[i, :], k2) + np.matmul(b3[i, :], k3) + np.matmul(b4[i, :], k4)

        gradbx = b1[i, 0]*k1 + b2[i, 0]*k2 + b3[i, 0]*k3 + b4[i, 0]*k4
        gradby = b1[i, 1]*k1 + b2[i, 1]*k2 + b3[i, 1]*k3 + b4[i, 1]*k4
        gradbz = b1[i, 2]*k1 + b2[i, 2]*k2 + b3[i, 2]*k3 + b4[i, 2]*k4

        barycentre = (p1[i, 0:3] + p2[i, 0:3] + p3[i, 0:3] + p4[i, 0:3])/4.0

        # and here is the field at the barycentre (calculate 4 ways)
        baryb[i, 0] = b1[i, 0] + np.sum(gradbx*(barycentre-p1[i, 0:3]))
        baryb[i, 1] = b1[i, 1] + np.sum(gradby*(barycentre-p1[i, 0:3]))
        baryb[i, 2] = b1[i, 2] + np.sum(gradbz*(barycentre-p1[i, 0:3]))

        baryb2[i, 0] = b2[i, 0] + np.sum(gradbx*(barycentre-p2[i, 0:3]))
        baryb2[i, 1] = b2[i, 1] + np.sum(gradby*(barycentre-p2[i, 0:3]))
        baryb2[i, 2] = b2[i, 2] + np.sum(gradbz*(barycentre-p2[i, 0:3]))

        baryb3[i, 0] = b3[i, 0] + np.sum(gradbx*(barycentre-p3[i, 0:3]))
        baryb3[i, 1] = b3[i, 1] + np.sum(gradby*(barycentre-p3[i, 0:3]))
        baryb3[i, 2] = b3[i, 2] + np.sum(gradbz*(barycentre-p3[i, 0:3]))

        baryb4[i, 0] = b4[i, 0] + np.sum(gradbx*(barycentre-p4[i, 0:3]))
        baryb4[i, 1] = b4[i, 1] + np.sum(gradby*(barycentre-p4[i, 0:3]))
        baryb4[i, 2] = b4[i, 2] + np.sum(gradbz*(barycentre-p4[i, 0:3]))

        # (these above all agree so this is the magnetic field at the barycentre)
        divb[i, 0] = time
        divb[i, 1] = divergence
        divb[i, 2] = curlmag[0]
        divb[i, 3] = curlmag[1]
        divb[i, 4] = curlmag[2]

        # the cross product of the calculated curl and the sample field times 1e-21 (SI), divided by m0

        # curl is in nT/km, nT/km*1e-12 = T/m
        # field is in nT, nT*1e-9 = T
        # j is curl B / m0 (curl B = m0*j)
        # use the magnetic field at the barycentre

        # compute the current components and total specifically
        jtotal[i, 0:3] = 1e-12*divb[i, 2:5]/m0
        jtotal[i, 3] = np.sqrt(jtotal[i, 0]**2+jtotal[i, 1]**2+jtotal[i, 2]**2)

        # compute the parallel and perpendicular components of the current
        btotal[i] = np.sqrt(np.dot(baryb[i, 0:3], baryb[i, 0:3]))

        # parallel is J.B/|B|
        jparallel[i] = np.dot(jtotal[i, 0:3], baryb[i, 0:3])/btotal[i]
        jparallel[i] = jparallel[i][0]

        # perp is J - J// B/|B| (components and total perpendicular current)
        jperpvec[i, 0:3] = jtotal[i, 0:3] - (jparallel[i]*baryb[i, 0:3])/btotal[i]
        jperpvec[i, 3] = np.sqrt(jperpvec[i, 0]**2 + jperpvec[i, 1]**2 + jperpvec[i, 2]**2)

        # alpha parameter
        alphaparallel[i] = np.abs(jparallel[i])/(1e-9*btotal[i])
        alpha[i] = np.abs(jtotal[i, 3])/(1e-9*btotal[i])

    # create the output variables
    store_data('baryb' + suffix, data={'x': timesb1, 'y': baryb})
    store_data('curlB' + suffix, data={'x': timesb1, 'y': divb[:, 2:5]})
    store_data('divB' + suffix, data={'x': timesb1, 'y': divb[:, 1]})
    store_data('jtotal' + suffix, data={'x': timesb1, 'y': jtotal[:, 0:3]})
    store_data('jpar' + suffix, data={'x': timesb1, 'y': jparallel})
    store_data('jperp' + suffix, data={'x': timesb1, 'y': jperpvec[:, 0:3]})
    store_data('alpha' + suffix, data={'x': timesb1, 'y': alpha})
    store_data('alphaparallel' + suffix, data={'x': timesb1, 'y': alphaparallel})

    # set some options
    options('baryb' + suffix, 'ytitle', 'baryb [nT]')
    options('divB' + suffix, 'ytitle', 'div(B) [nT/km]')
    options('curlB' + suffix, 'ytitle', 'curl(B) [nT/km]')
    options('curlB' + suffix, 'Color', ['b', 'g', 'r'])
    options('curlB' + suffix, 'legend_names', ['delBx', 'delBy', 'delBz'])
    options('jtotal' + suffix, 'ytitle', 'J [A/m^2]')
    options('jtotal' + suffix, 'Color', ['b', 'g', 'r'])
    options('jtotal' + suffix, 'legend_names', ['Jx', 'Jy', 'Jz'])
    options('jperp' + suffix, 'ytitle', 'Jperp [A/m^2]')
    options('jperp' + suffix, 'Color', ['b', 'g', 'r'])
    options('jperp' + suffix, 'legend_names', ['Jperpx', 'Jperpy', 'Jperpz'])
    options('jpar' + suffix, 'ytitle', 'Jparallel [A/m^2]')

    return ['baryb', 'curlB', 'divB', 'jtotal', 'jpar', 'jperp', 'alpha', 'alphaparallel']



import numpy as np
from pyspedas import tinterpol, tdeflag
from pyspedas.geopack.get_w_params import get_w
from pytplot import get_data, store_data

def get_tsy_params(dst_tvar, imf_tvar, Np_tvar, Vp_tvar, model, pressure_tvar=None, newname=None, speed=False, g_variables=None):
    """
    This procedure will interpolate inputs, generate
    Tsyganenko model parameters and store them in a tplot 
    variable that can be passed directly to the model 
    procedure.

    Input
    ------
        dst_tvar: str
            tplot variable containing the Dst index

        imf_tvar: str
            tplot variable containing the interplanetary
            magnetic field vector in GSM coordinates

        Np_tvar: str
            tplot variable containing the solar wind 
            ion density (cm**-3)

        Vp_tvar: str
            tplot variable containing the proton velocity

        model: str
            Tsyganenko model; should be: 'T89', T96', 'T01','TS04'

    Parameters
    -----------
        newname: str
            name of the output variable; default: t96_par,
            't01_par' or 'ts04_par', depending on the 
            model

        speed: bool
            Flag to indicate Vp_tvar is speed, and not velocity
            (defaults to False)
        
        pressure_tvar: str
            Set this to specify a tplot variable containing 
            solar wind dynamic pressure data. If not supplied, 
            it will be calculated internally from proton density 
            and proton speed.

    Returns
    --------
        Name of the tplot variable containing the parameters. 

        The parameters are:
            (1) solar wind pressure pdyn (nanopascals),
            (2) dst (nanotesla),
            (3) byimf,
            (4) bzimf (nanotesla)
            (5-10) indices w1 - w6, calculated as time integrals from the beginning of a storm
                see the reference (3) below, for a detailed definition of those variables

    """
    model = model.lower()

    if model not in ['t89', 't96', 't01', 'ts04']:
        print('Unknown model: ' + model)
        return

    tdeflag(Np_tvar, method='remove_nan', overwrite=True)
    tdeflag(dst_tvar, method='remove_nan', overwrite=True)
    tdeflag(Vp_tvar, method='remove_nan', overwrite=True)

    # interpolate the inputs to the Np timestamps
    tinterpol(imf_tvar, Np_tvar, newname=imf_tvar+'_interp')
    tinterpol(dst_tvar, Np_tvar, newname=dst_tvar+'_interp')
    tinterpol(Vp_tvar, Np_tvar, newname=Vp_tvar+'_interp')

    if pressure_tvar is not None:
        tdeflag(pressure_tvar, method='remove_nan', overwrite=True)
        tinterpol(pressure_tvar, Np_tvar, newname=pressure_tvar+'_interp')

    Np_data = get_data(Np_tvar)
    dst_data = get_data(dst_tvar+'_interp')
    imf_data = get_data(imf_tvar+'_interp')
    Vp_data = get_data(Vp_tvar+'_interp')

    if pressure_tvar is not None:
        P_data = get_data(pressure_tvar+'_interp')

    if model == 't96':
        out = np.array((P_data.y, 
                        dst_data.y, 
                        imf_data.y[:, 1], 
                        imf_data.y[:, 2], 
                        np.zeros(len(dst_data.y)), 
                        np.zeros(len(dst_data.y)), 
                        np.zeros(len(dst_data.y)), 
                        np.zeros(len(dst_data.y)), 
                        np.zeros(len(dst_data.y)), 
                        np.zeros(len(dst_data.y))))
    elif model == 't01':
        if g_variables is None:
            print('G variables required for T01 model; create a tplot variable containing the G variables, and provide the name of that keyword to the g_variables keyword.')
            return
        else:
            if isinstance(g_variables, str):
                g_data = get_data(g_variables)

                if g_data is None:
                    print('Problem reading G variable: ' + g_variables)
                    return

                g1 = g_data.y[:, 0]
                g2 = g_data.y[:, 1]
            else:
                if isinstance(g_variables, list):
                    g_variables = np.array(g_variables)
                    
                if len(g_variables.shape) > 1:
                    g1 = g_variables[:, 0]
                    g2 = g_variables[:, 1]
                else:
                    g1 = np.repeat(g_variables[0], len(dst_data.y))
                    g2 = np.repeat(g_variables[1], len(dst_data.y))

            out = np.array((P_data.y, 
                            dst_data.y, 
                            imf_data.y[:, 1], 
                            imf_data.y[:, 2], 
                            g1, 
                            g2, 
                            np.zeros(len(dst_data.y)), 
                            np.zeros(len(dst_data.y)), 
                            np.zeros(len(dst_data.y)), 
                            np.zeros(len(dst_data.y))))
    elif model == 'ts04':
        params = get_w(trange=[np.nanmin(Np_data.times), np.nanmax(Np_data.times)], create_tvar=True)
        # interpolate the inputs to the Np timestamps
        tinterpol(params, Np_tvar, newname=params+'_interp')
        w_data = get_data(params+'_interp')

        if w_data is None:
            print('Problem loading W variables for TS04 model.')
            return

        out = np.array((P_data.y, 
                        dst_data.y, 
                        imf_data.y[:, 1], 
                        imf_data.y[:, 2], 
                        w_data.y[:, 0], 
                        w_data.y[:, 1], 
                        w_data.y[:, 2], 
                        w_data.y[:, 3], 
                        w_data.y[:, 4], 
                        w_data.y[:, 5]))
    elif model == 't01':
        print('not implemented yet')
        return

    if newname is None:
        newname = model + '_par'

    saved = store_data(newname, data={'x': dst_data.times, 'y': out.T})

    if saved:
        return newname






from pyspedas import time_double
from pyspedas.mms.mms_load_fast_segments import mms_load_fast_segments
from pyspedas.mms.mms_load_sroi_segments import mms_load_sroi_segments
from pyspedas.mms.mms_load_brst_segments import mms_load_brst_segments

def spd_mms_load_bss(trange=['2015-10-16', '2015-10-17'], datatype=['fast', 'burst'], 
                     include_labels=False, probe='1', nodownload=False):
    """
    Creates tplot variables which allow you to display horizontal color bars 
    indicating burst data availability.

    Parameters
    -----------
        trange: list of str
            Time frame for the bars

        datatype: str or list of str
            Type of BSS data (current valid options: 'fast', 'burst')

        include_labels: bool
            Flag to have horizontal bars labeled

        probe: str or int
            S/C probe # for SRoI bars (used as fast survey segments after 6Nov15; default: 1)
    """

    if not isinstance(datatype, list):
        datatype = [datatype]

    if include_labels:
        burst_label = 'Burst'
        fast_label = 'Fast'
    else:
        burst_label = ''
        fast_label = ''

    for dtype in datatype:
        if dtype == 'fast':
            if time_double(trange[0]) <= time_double('2015-11-06'):
                # use the old fast segments code for dates before 6Nov15
                out = mms_load_fast_segments(trange=trange)
            else:
                # use SRoI code for dates on and after 6Nov15
                out = mms_load_sroi_segments(trange=trange, probe=probe)
        elif dtype == 'burst':
            out = mms_load_brst_segments(trange=trange)
        else:
            print('Unsupported datatype: ' + dtype + '; valid options: "fast" and "burst"')
            continue

        if out is None:
            print('Problem loading segments for ' + dtype)




def spd_units_string(units, units_only=False):
    """
    Return string describing particle data units for labels

    Input:
        units: str
            String describing units

    Returns:
        String containing full units string
    """
    out = ['', 'Unknown', '']
    units = units.lower()

    if units == 'counts':
        out = ['', 'Counts', '']
    elif units == 'rate':
        out = ['Rate (', '#/sec', ')']
    elif units == 'eflux':
        out = ['Energy Flux (', 'eV / sec / cm^2 / ster / eV', ')']
    elif units == 'flux':
        out = ['Flux (', '# / sec / cm^2 / ster / eV', ')']
    elif units == 'df':
        out = ['f (', 's^3 / cm^3 / km^3', ')']
    elif units == 'df_cm':
        out = ['f (', 's^3 / cm^6', ')']
    elif units == 'df_km':
        out = ['f (', 's^3 / km^6', ')']
    elif units == 'e2flux':
        out = ['Energy^2 Flux (', 'eV^2 / sec / cm^2 / ster /eV', ')']
    elif units == 'e3flux':
        out = ['Energy^3 Flux (', 'eV^3 / sec / cm^2 / ster /eV', ')']

    if units_only:
        return out[1]

    return ''.join(out)


from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.analysis.time_clip import time_clip as tclip
from pytplot import netcdf_to_tplot

from .config import CONFIG

def load(trange=['2013-11-5', '2013-11-6'], 
         probe='15',
         instrument='fgm',
         datatype='1min', 
         suffix='', 
         downloadonly=False,
         no_update=False,
         time_clip=False):
    """
    This function loads data from the GOES mission; this function is not meant 
    to be called directly; instead, see the wrappers:
        pyspedas.goes.fgm
        pyspedas.goes.eps
        pyspedas.goes.epead
        pyspedas.goes.maged
        pyspedas.goes.magpd
        pyspedas.goes.hepad
        pyspedas.goes.xrs

    """

    if not isinstance(probe, list):
        probe = [probe]

    fullavgpath = ['full', 'avg']
    goes_path_dir = fullavgpath[datatype == '1min' or datatype == '5min']

    for prb in probe:
        remote_path = goes_path_dir + '/%Y/%m/goes' + str(prb) + '/netcdf/'

        if instrument == 'fgm':
            if datatype == '512ms': # full, unaveraged data
                pathformat = remote_path + 'g' + str(prb) + '_magneto_512ms_%Y%m%d_%Y%m%d.nc'
            elif datatype == '1min': # 1 min averages
                pathformat = remote_path + 'g' + str(prb) + '_magneto_1m_%Y%m01_%Y%m??.nc'
            elif datatype == '5min': # 5 min averages
                pathformat = remote_path + 'g' + str(prb) + '_magneto_5m_%Y%m01_%Y%m??.nc'
        elif instrument == 'eps':
            # energetic particle sensor -- only valid for GOES-08 through GOES-12, only averaged data available
            if datatype == '1min':
                pathformat = remote_path + 'g' + str(prb) + '_eps_1m_%Y%m01_%Y%m??.nc'
            else:
                pathformat = remote_path + 'g' + str(prb) + '_eps_5m_%Y%m01_%Y%m??.nc'
        elif instrument == 'epead':
            # electron, proton, alpha detector -- only valid on GOES-13, 14, 15
            if datatype == '1min':
                pathformat = [remote_path + 'g' + str(prb) + '_epead_e13ew_1m_%Y%m01_%Y%m??.nc',
                                                             '_epead_p17ew_1m_%Y%m01_%Y%m??.c',
                                                             '_epead_a16ew_1m_%Y%m01_%Y%m??.nc']
            elif datatype == '5min':
                pathformat = [remote_path + 'g' + str(prb) + '_epead_e13ew_5m_%Y%m01_%Y%m??.nc',
                                                             '_epead_p17ew_5m_%Y%m01_%Y%m??.c',
                                                             '_epead_a16ew_5m_%Y%m01_%Y%m??.nc']
            else:
                pathformat = [remote_path + 'g' + str(prb) + '_epead_e1ew_4s_%Y%m%d_%Y%m%d.nc',
                                                             '_epead_e2ew_16s_%Y%m%d_%Y%m%d.nc',
                                                             '_epead_e3ew_16s_%Y%m%d_%Y%m%d.nc',
                                                             '_epead_p1ew_8s_%Y%m%d_%Y%m%d.nc',
                                                             '_epead_p27e_32s_%Y%m%d_%Y%m%d.nc',
                                                             '_epead_p27w_32s_%Y%m%d_%Y%m%d.nc',
                                                             '_epead_a16e_32s_%Y%m%d_%Y%m%d.nc',
                                                             '_epead_a16w_32s_%Y%m%d_%Y%m%d.nc']
        elif instrument == 'maged':
            # magnetospheric electron detector -- only valid on GOES 13, 14, 15
            if datatype == '1min':
                pathformat = remote_path + 'g' + str(prb) + '_maged_19me15_1m_%Y%m01_%Y%m??.nc'
            elif datatype == '5min':
                pathformat = remote_path + 'g' + str(prb) + '_maged_19me15_5m_%Y%m01_%Y%m??.nc'
            else:
                channels = ['me1','me2','me3','me4','me5']
                resolution = ['2','2','4','16','32']
                pathformat = []
                for idx, channel in enumerate(channels):
                    pathformat.append(remote_path + 'g' + str(prb) + '_maged_19'+channel+'_'+resolution[idx]+'s_%Y%m%d_%Y%m%d.nc')
        elif instrument == 'magpd':
            # magnetospheric proton detector -- only valid on GOES 13, 14, 15
            if datatype == '1min':
                pathformat = remote_path + 'g' + str(prb) + '_magpd_19mp15_1m_%Y%m01_%Y%m??.nc'
            elif datatype == '5min':
                pathformat = remote_path + 'g' + str(prb) + '_magpd_19mp15_5m_%Y%m01_%Y%m??.nc'
            else:
                channels = ['mp1','mp2','mp3','mp4','mp5']
                resolution = ['16','16','16','32','32']
                pathformat = []
                for idx, channel in enumerate(channels):
                    pathformat.append(remote_path + 'g' + str(prb) + '_magpd_19'+channel+'_'+resolution[idx]+'s_%Y%m%d_%Y%m%d.nc')
        elif instrument == 'hepad':
            # high energy proton and alpha detector -- valid for GOES 08-15
            if datatype == '1min':
                pathformat = [remote_path + 'g' + str(prb) + '_hepad_ap_1m_%Y%m01_%Y%m??.nc',
                                                             '_hepad_s15_1m_%Y%m01_%Y%m??.nc']
            elif datatype == '5min':
                pathformat = [remote_path + 'g' + str(prb) + '_hepad_ap_5m_%Y%m01_%Y%m??.nc',
                                                             '_hepad_s15_5m_%Y%m01_%Y%m??.nc']
            else:
                pathformat = [remote_path + 'g' + str(prb) + '_hepad_ap_32s_%Y%m%d_%Y%m%d.nc',
                                                             '_hepad_s15_4s_%Y%m%d_%Y%m%d.nc']
        elif instrument == 'xrs':
            # x-ray sensor -- valid for GOES 08-15
            if datatype == '1min':
                pathformat = remote_path + 'g' + str(prb) + '_xrs_1m_%Y%m01_%Y%m??.nc'
            elif datatype == '5min':
                pathformat = remote_path + 'g' + str(prb) + '_xrs_5m_%Y%m01_%Y%m??.nc'
            else:
                pathformat = remote_path + 'g' + str(prb) + '_xrs_2s_%Y%m%d_%Y%m%d.nc'

        # find the full remote path names using the trange
        if isinstance(pathformat, list):
            remote_names = []
            for path in pathformat:
                remote_names.extend(dailynames(file_format=path, trange=trange))
        else:
            remote_names = dailynames(file_format=pathformat, trange=trange)

        out_files = []

        files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update)
        if files is not None:
            for file in files:
                out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = netcdf_to_tplot(out_files, suffix=suffix, merge=True, time='time_tag')

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



import warnings
import numpy as np
from pytplot import get_data, store_data, options

# use nanmean from bottleneck if it's installed, otherwise use the numpy one
# bottleneck nanmean is ~2.5x faster
try:
    import bottleneck as bn
    nanmean = bn.nanmean
except ImportError:
    nanmean = np.nanmean

def mms_feeps_spin_avg(probe='1', data_units='intensity', datatype='electron', data_rate='srvy', level='l2', suffix=''):
    """
    This function will spin-average the omni-directional FEEPS energy spectra
    
    Parameters:
        probe: str
            probe #, e.g., '4' for MMS4

        data_units: str
            'intensity' or 'count_rate'

        datatype: str
            'electron' or 'ion'

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst'

        level: str
            data level, e.g., 'l2'
            
        suffix: str
            suffix of the loaded data

    Returns:
        Name of tplot variable created.
    """

    units_label = ''
    if data_units == 'intensity':
        units_label = '1/(cm^2-sr-s-keV)'
    elif data_units == 'counts':
        units_label = '[counts/s]'

    if datatype == 'electron':
        lower_en = 71.0
    else:
        lower_en = 78.0

    prefix = 'mms'+str(probe)+'_epd_feeps_'

    # get the spin sectors
    # v5.5+ = mms1_epd_feeps_srvy_l1b_electron_spinsectnum
    sector_times, spin_sectors = get_data(prefix + data_rate + '_' + level + '_' + datatype + '_spinsectnum' + suffix)

    spin_starts = [spin_end + 1 for spin_end in np.where(spin_sectors[:-1] >= spin_sectors[1:])[0]]

    var_name = prefix + data_rate + '_' + level + '_' + datatype + '_' + data_units + '_omni'

    times, data, energies = get_data(var_name + suffix)

    spin_avg_flux = np.zeros([len(spin_starts), len(energies)])

    current_start = spin_starts[0]
    for spin_idx in range(1, len(spin_starts)-1):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            spin_avg_flux[spin_idx-1, :] = nanmean(data[current_start:spin_starts[spin_idx]+1, :], axis=0)
        current_start = spin_starts[spin_idx] + 1

    store_data(var_name + '_spin' + suffix, data={'x': times[spin_starts], 'y': spin_avg_flux, 'v': energies})
    options(var_name + '_spin' + suffix, 'spec', True)
    options(var_name + '_spin' + suffix, 'ylog', True)
    options(var_name + '_spin' + suffix, 'zlog', True)
    options(var_name + '_spin' + suffix, 'yrange', [lower_en, 600.0])
    options(var_name + '_spin' + suffix, 'Colormap', 'spedas')
    options(var_name + '_spin' + suffix, 'ztitle', units_label)
    options(var_name + '_spin' + suffix, 'ytitle', 'MMS' + str(probe) + ' ' + datatype + ' (keV)')

    return var_name + '_spin' + suffix




import os
import unittest
from pyspedas.utilities.data_exists import data_exists

import pyspedas

class LoadTestCases(unittest.TestCase):
    def test_downloadonly(self):
        mag_files = pyspedas.goes.fgm(datatype='1min', downloadonly=True)
        self.assertTrue(os.path.exists(mag_files[0]))

    def test_load_1min_mag_data(self):
        mag_vars = pyspedas.goes.fgm(datatype='1min')
        self.assertTrue(data_exists('BX_1'))
        self.assertTrue(data_exists('BY_1'))
        self.assertTrue(data_exists('BZ_1'))

    def test_load_mag_data(self):
        mag_vars = pyspedas.goes.fgm(datatype='512ms', suffix='_512')
        self.assertTrue(data_exists('BX_1_512'))
        self.assertTrue(data_exists('BY_1_512'))
        self.assertTrue(data_exists('BZ_1_512'))

if __name__ == '__main__':
    unittest.main()


"""Test gmag and themis load functions."""
import os
import unittest
from pyspedas.utilities.data_exists import data_exists

import pyspedas


class GmagTestCases(unittest.TestCase):
    """Test GMAG functions."""

    def test_get_group(self):
        """Get gmag stations of a group."""
        from pyspedas.themis.ground.gmag import get_group
        self.assertTrue(get_group('ccnv') == ['epo'])

    def test_gmag_list(self):
        """Get gmag list of stations."""
        from pyspedas.themis.ground.gmag import gmag_list
        self.assertTrue(gmag_list()[0:5] == ['abk', 'akul', 'amd', 'amer',
                                             'amk'])

    def test_gmag_groups(self):
        """Get gmag groups."""
        from pyspedas.themis.ground.gmag import gmag_groups
        gmag_table = gmag_groups()
        self.assertTrue(list(gmag_table.keys())[0:5] == ['kyoto', 'sgu',
                                                         'autx', 'ae', 'aari'])

    def test_check_gmag(self):
        """Check a gmag station."""
        from pyspedas.themis.ground.gmag import check_gmag
        self.assertTrue(check_gmag('ccnv') == 1)
        self.assertTrue(check_gmag('abcd') == 0)

    def test_load_gmag_data(self):
        """Load gmag."""
        pyspedas.themis.gmag(varnames=['thg_mag_amer'])
        self.assertTrue(data_exists('thg_mag_amer'))


class LoadTestCases(unittest.TestCase):
    """Test themis load functions."""

    def test_load_state_data(self):
        """Load state."""
        pyspedas.themis.state(varnames=['thc_pos'])
        self.assertTrue(data_exists('thc_pos'))

    def test_load_sst_data(self):
        """Load SST."""
        pyspedas.themis.sst(varnames=['thc_psif_en_eflux'])
        self.assertTrue(data_exists('thc_psif_en_eflux'))

    def test_load_fgm_data(self):
        """Load FGM."""
        pyspedas.themis.fgm(varnames=['thc_fgs_btotal'])
        self.assertTrue(data_exists('thc_fgs_btotal'))

    def test_load_fit_data(self):
        """Load FIT."""
        pyspedas.themis.fit(varnames=['thc_fgs_gse'])
        self.assertTrue(data_exists('thc_fgs_gse'))

    def test_load_esa_data(self):
        """Load ESA."""
        pyspedas.themis.esa(varnames=['thc_peif_density'])
        self.assertTrue(data_exists('thc_peif_density'))

    def test_load_fft_data(self):
        """Load FFT."""
        pyspedas.themis.fft(varnames=['thc_ffp_16_edc34'])
        self.assertTrue(data_exists('thc_ffp_16_edc34'))

    def test_load_fft_l1_data(self):
        """Load L1 FFT."""
        pyspedas.themis.fft(level='l1', varnames=['thc_ffp_16'])
        self.assertTrue(data_exists('thc_ffp_16'))

    def test_load_fbk_data(self):
        """Load FBK."""
        pyspedas.themis.fbk(varnames=['thc_fb_hff'])
        self.assertTrue(data_exists('thc_fb_hff'))

    def test_load_mom_data(self):
        """Load MOM."""
        pyspedas.themis.mom(varnames=['thc_peim_density'])
        self.assertTrue(data_exists('thc_peim_density'))

    def test_load_gmom_data(self):
        """Load GMOM."""
        pyspedas.themis.gmom(trange=['2020-01-01', '2020-01-01'],
                             varnames=['thc_ptiff_density'])
        self.assertTrue(data_exists('thc_ptiff_density'))

    def test_load_scm_data(self):
        """Load SCM."""
        pyspedas.themis.scm(varnames=['thc_scf_btotal'])
        self.assertTrue(data_exists('thc_scf_btotal'))

    def test_load_scm_l1_data(self):
        """Load L1 SCM."""
        pyspedas.themis.scm(level='l1', varnames=['thc_scf'])
        self.assertTrue(data_exists('thc_scf'))

    def test_load_efi_l1_data(self):
        """Load L1 EFI."""
        pyspedas.themis.efi(level='l1', varnames=['thc_eff'])
        self.assertTrue(data_exists('thc_eff'))

    def test_load_efi_data(self):
        """Load EFI."""
        pyspedas.themis.efi(time_clip=True, varnames=['thc_eff_e12_efs'])
        self.assertTrue(data_exists('thc_eff_e12_efs'))

    def test_downloadonly(self):
        """Downloadonly keyword."""
        files = pyspedas.themis.efi(downloadonly=True,
                                    trange=['2014-2-15', '2014-2-16'],
                                    varnames=['thc_eff_e12_efs'])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()



from .satellite.erg.hep.hep import hep
from .satellite.erg.lepe.lepe import lepe
from .satellite.erg.lepi.lepi import lepi
from .satellite.erg.mepe.mepe import mepe
from .satellite.erg.mepi.mepi_nml import mepi_nml
from .satellite.erg.mepi.mepi_tof import mepi_tof
from .satellite.erg.mgf.mgf import mgf
from .satellite.erg.orb.orb import orb
from .satellite.erg.pwe.pwe_efd import pwe_efd
from .satellite.erg.pwe.pwe_hfa import pwe_hfa
from .satellite.erg.pwe.pwe_ofa import pwe_ofa
from .satellite.erg.pwe.pwe_wfc import pwe_wfc
from .satellite.erg.xep.xep import xep